/* See {dm_datum.h}. */
/* Last edited on 2006-12-25 18:57:42 by stolfi */

#define dm_datum_C_COPYRIGHT \
  "Copyright  2006  by the State University of Campinas (UNICAMP)"

#include <dm_basic.h>
#include <dm_nucleic.h>
#include <dm_sample.h>
#include <dm_datum.h>
#include <dm_seq.h>

#include <vec.h>
#include <affirm.h>

#include <math.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>

#include <math.h>
#include <ctype.h>
#include <assert.h>

/* IMPLEMENTATIONS: */

/* QUEST OF THE CORRECT DATUM DISTANCE ESTIMATORS

  Suppose we have two datums {X,Y}, each of them approximately
  equal to the weighted average of several discrete datums {x[0..N-1]}
  and {y[0..N-1]}:
  
    {X ~ H(x) = SUM{ w[i]*x[i] : i = 0..N-1}}
    {Y ~ H(y) = SUM{ w[i]*y[i] : i = 0..N-1}}
    
  where {w[0..N-1]} are real weights.
  
  We wish to estimate the random variable 
  
    {S2(x,y) = SUM{ a[i]*|x[i] - y[i]|^2 : i = 0..N-1 }}
  
  for some set of weights {a[0..N-1]} (possibly related to the {w}
  weights), assuming that the datums {x[0..N-1]} and {y[0..N-1]} are
  taken from some fixed distribution.. Namely, we wish to compute
  
    {D2(X,Y) = E(S2(x,y) @ H(x)~X & H(y)~Y)}
    
  where {E(A @ B)} is the expectation of random variable {A}
  gien condition {B}.  By the expectation-of-sum theorem have
  
    {D2(X,Y) = SUM{ a[i] * F2[i](X,Y): i = 0..N-1 }}
    
  where
    
    {F2[i](X,Y) = E(|x[i] - y[i]|^2 @ H(x)~X & H(y)~Y)}
    
  In our case, the vectors {x[i]} and {y[i]} are taken from the set
  {V} of the vertices of a regular tetrahedron whose edge has length
  {sqrt(8)}. So {|x[i] - y[i]|^2} is 0 if {x[i] = y[i]}, and 8 if
  {x[i] != y[i]}. So
  
    {F2[i](X,Y) = 8 * (1 - Q[i](X,Y))}
    
  where
  
    {Q[i](X,Y) = Pr(x[i] = y[i] @ H(x)~X & H(y)~Y)}
  
  Assuming that, /a priori/, the choice of
  {x[i]} is indepedent of {y[0..N-1]}, and vice-versa,
  we have
  
    {Q[i](X,Y) = SUM{ Pr(x[i]=u @ H(x)~X)*Pr(y[i]=u @ H(y)~Y) : u in V }}
  
  So we need to estimate 
  
    {G[i](X,u) = Pr(x[i]=u @ H(x)~X)}
  
  Good question...
  
  ----------------------------------------------------------------------
  
  Another approach is to look for a second-degree approximation to the
  function {D2(X,Y)}. The empirical approach is as follows: choose
  weights {w[0..N-1]} and {a[0..N-1]}. Generate a large number of test
  cases {x[0..N-1]} and {y[0..N-1]}.  For each case, compute {X=H(x)},
  {Y=h(y)}, and {Z=S2(x,y)}.  Now find a degree-2 function {Z~T2(X,Y)}
  that provides the best fit to all those data points. 
  
  The function {T2} should be symmetric under arbitrary channel
  permutations and exchange of {X} with {Y}. Therefore it must be a
  linear combination of the homogeneous functions of degree {0,1,2}
  with those symmetries:
  
    {L0(X,Y) = 1}
    {L1(X,Y) = X0 + X1 + X2 + Y0 + Y1 + Y2}
    {L2(X,Y) = X0^2 + X1^2 + X2^2 + Y0^2 + Y1^2 + Y2^2}
    {L3(X,Y) = X0*X1 + X1*X2 + X2*X0 + Y0*Y1 + Y1*Y2 + Y2*Y0}
    {L4(X,Y) = X0*Y0 + X1*Y1 + X2*Y2}
    {L5(X,Y) = X0*Y1 + X1*Y2 + X2*Y0 + X1*Y0 + X2*Y1 + X0*y2}
    
  Moreover, the function {T2} should be unchanged by simultaneous 
  negation of {X0,X1,Y0,Y1}.  That excludes {L1,L3,L5}, leaving only
  {L0,L2,L4}. 
    
  Another equivalent basis is {1,<X|Y>,(|X|^2+|Y|^2)} where {<X|Y>}
  is the dot product of the two vectors.
  
  Numerical experiments indicate that the optimum quadratic distance
  estimators is {T2(X,Y) = 6 - K*<X|Y>} where {K} is exactly 2 for uniform 
  weights ({w[i] = 1/N} for all {i}) and somewhat higher than 2
  for other weights.  For binomial weights {w[i] = choose(N-1,i)/2^N}
  the optimal {K} is given by the table
   
     { N =     1     2     3     4     5     6     7  } 
     {     ----- ----- ----- ----- ----- ----- -----  }
     { K = 2.000 2.000 2.222 2.242 2.255 2.269 2.277  }
  
  */

#define dm_DIFF2_K 2.000
  /* The coefficient {K} in the estimation formula {T2(X,Y) = 6 - K*<X|Y>}
    for {S2(x,y)}, given {X=H(x)} and {Y=H(y)}.
    !!! Should it be closer to 2.25? */

double dm_datum_diffsq(dm_datum_t *xd, dm_datum_t *yd)
  { /* Given the above discussion, we use the formula 
      {S2(x,y) ~ (6 - K*<X|Y>)/8}. */
    
    double dprod = 0;
    int c;
    for (c = 0; c < dm_CHANNELS; c++)
      { double sx = dm_sample_decode(xd->c[c]);
        double sy = dm_sample_decode(yd->c[c]);
        dprod += sx*sy;
      }
    double T2 = (6 - dm_DIFF2_K*dprod)/8;
    if (T2 < 0) { T2 = 0; }
    return T2;
  }

double dm_datum_step_diffsq
  ( dm_datum_t *gxd, 
    dm_datum_t *gyd, 
    dm_datum_t *hxd, 
    dm_datum_t *hyd
  )
  { double dprod = 0;
    int c;
    for (c = 0; c < dm_CHANNELS; c++)
      { double sxg = dm_sample_decode(gxd->c[c]);
        double syg = dm_sample_decode(gyd->c[c]);
        double sxh = dm_sample_decode(hxd->c[c]);
        double syh = dm_sample_decode(hyd->c[c]);
        dprod += (2*sxg*syg + sxg*syh + sxh*syg + 2*sxh*syh)/6;
      }
    double T2 = (6 - dm_DIFF2_K*dprod)/8;
    if (T2 < 0) { T2 = 0; }
    return T2;
  }

void dm_datum_to_nucleic_densities(dm_datum_t *d, double *A, double *T, double *C, double *G)
  { double scale = dm_NUCLEIC_RAW_SCALE;
    double d0 = dm_sample_decode(d->c[0])*scale;
    double d1 = dm_sample_decode(d->c[1])*scale;
    double d2 = dm_sample_decode(d->c[2])*scale;
    (*A) = ((+d0) + (+d1) + (+d2) + 1)/4;
    (*T) = ((-d0) + (-d1) + (+d2) + 1)/4;
    (*C) = ((+d0) + (-d1) + (-d2) + 1)/4;
    (*G) = ((-d0) + (+d1) + (-d2) + 1)/4;
  }

void dm_datum_decoded_from_nucleic_char(char b, int *d)
  { int A, T, C, G;
    dm_nucleic_value(b, &A, &T, &C, &G);
    d[0] = (A-T)+(C-G);
    d[1] = (G-C)+(A-T);
    d[2] = (A+T)-(C+G);
  }

dm_datum_t dm_datum_encoded_from_nucleic_char(char b)
  { int d[3];
    dm_datum_decoded_from_nucleic_char(b, d);
    dm_datum_t r;
    double scale = dm_NUCLEIC_RAW_SCALE;
    r.c[0] = dm_sample_encode(((double)d[0])/scale);
    r.c[1] = dm_sample_encode(((double)d[1])/scale);
    r.c[2] = dm_sample_encode(((double)d[2])/scale);
    return r;
  }

dm_datum_vec_t dm_datum_vec_from_nucleic_string(char *s)
  { /* Get count {nbas} of nucleotide characters: */
    int nbas = 0; 
    char *p = s;
    while ((*p) != 0) { if (is_dna_basis(*p)) { nbas++; } p++; }
    /* Allocate datum vector: */
    dm_datum_vec_t dv = dm_datum_vec_new(nbas);
    /* Convert to numeric and encode: */
    int j = 0;
    p = s;
    while ((*p) != 0)
      { char c = *p;
        if (is_dna_basis(c)) 
          { dv.el[j] = dm_datum_encoded_from_nucleic_char(c);
            j++;
          }
        p++;
      }
    assert(j == nbas);
    assert(dv.nel == nbas);
    return dv;
  }

vec_typeimpl(dm_datum_vec_t,dm_datum_vec,dm_datum_t);

void dm_datum_encoded_write(FILE *wr, dm_datum_t *d, char *lp, char *sep, char *rp)
{ if (lp != NULL) { fputs(lp, wr); }
    int c;
    for (c = 0; c < dm_CHANNELS; c++)
      { if ((c != 0) && (sep != NULL)) { fputs(sep, wr); }
        fprintf(wr, "%+4d", d->c[c]); 
      } 
    if (rp != NULL) { fputs(rp, wr); }
  }
    
void dm_datum_decoded_write(FILE *wr, dm_datum_t *d, char *lp, char *sep, char *rp)
  { if (lp != NULL) { fputs(lp, wr); }
    int c;
    for (c = 0; c < dm_CHANNELS; c++)
      { if ((c != 0) && (sep != NULL)) { fputs(sep, wr); }
        fprintf(wr, "%+7.4f", dm_sample_decode(d->c[c])); 
      } 
    if (rp != NULL) { fputs(rp, wr); }
  }
