/* See {dm_dyn.h} */
/* Last edited on 2006-12-25 20:22:34 by stolfi */

#define dm_dyn_C_COPYRIGHT \
  "Copyright  2005  by the State University of Campinas (UNICAMP)" \
  " and the Federal Fluminense University (UFF)"

#include <dm_dyn.h>
#include <dm_basic.h>
#include <dm_rung.h>
#include <dm_image.h>

#include <vec.h>
#include <affirm.h>
#include <float_image.h>

#include <stdint.h>
#include <assert.h>
#include <stdio.h>
#include <math.h>

vec_typeimpl(dm_dyn_entry_vec_t,dm_dyn_entry_vec,dm_dyn_entry_t);
 
dm_dyn_tableau_t dm_dyn_tableau_new(void)
  { dm_dyn_tableau_t tb;
    tb.minr = 0;
    tb.maxr = -1;
    tb.nr = 0;
    tb.mins = int_vec_new(0);
    tb.maxs = int_vec_new(0);
    tb.ns = 0;
    tb.ev = dm_dyn_entry_vec_new(0);
    return tb;
  }
    
void dm_dyn_tableau_resize(dm_dyn_tableau_t *tb, int minr, int maxr, int maxds)
  { demand(minr <= maxr, "bad {r} range");
    tb->minr = minr;
    tb->maxr = maxr;
    tb->nr = maxr - minr + 1;
    int_vec_expand(&(tb->mins), tb->nr - 1);
    int_vec_expand(&(tb->maxs), tb->nr - 1);
    int dr;
    for (dr = 0; dr < tb->nr; dr++) { tb->mins.el[dr] = 0; tb->maxs.el[dr] = -1; }
    demand(maxds % 2 == 0, "maxds should be even");
    tb->ns = maxds/2 + 1;
    dm_dyn_entry_vec_expand(&(tb->ev), tb->nr*tb->ns - 1);
  }
    
void dm_dyn_tableau_set_s_range(dm_dyn_tableau_t *tb, int r, int mins, int maxs)
  { demand((tb->minr <= r) && (r <= tb->maxr), "invalid {r}");
    int dr = r - tb->minr;
    demand((mins + r) % 2 == 0, "{mins} has wrong parity");
    demand((maxs + r) % 2 == 0, "{maxs} has wrong parity");
    demand((maxs - mins)/2 < tb->ns, "{s} range too large");
    tb->mins.el[dr] = mins;
    tb->maxs.el[dr] = maxs;
    int ds;
    for (ds = 0; ds < tb->ns; ds++)
      { int k = dr + ds*tb->nr;
        tb->ev.el[k] = (dm_dyn_entry_t){ -INF, dm_rung_none };
      }
  }
    
dm_dyn_entry_t *dm_dyn_tableau_get_entry_address(dm_dyn_tableau_t *tb, dm_rung_t g)
  { /* Convert {g} to the R,S coordinate system: */
    int r = g.c[0] + g.c[1];
    int s = g.c[0] - g.c[1];
    /* Find index {dr} into {mins,maxs}: */
    if ((r < tb->minr) || (r > tb->maxr)) { return NULL; }
    int dr = r - tb->minr;
    /* Get relative S-coordinate {ds}: */
    int mins = tb->mins.el[dr];
    int maxs = tb->maxs.el[dr];
    if ((s < mins) || (s > maxs)) { return NULL; }
    assert((s - mins) % 2 == 0);
    int ds = (s - mins)/2;
    int k = dr + ds*tb->nr;
    return &(tb->ev.el[k]);
  }

void dm_dyn_tableau_free(dm_dyn_tableau_t *tb)
  { free(tb->mins.el);
    free(tb->maxs.el);
    free(tb->ev.el);
  }

void dm_dyn_tableau_image_write_named
  ( int nx,               /* Length of first sequence. */
    int ny,               /* Length of second sequence. */
    dm_dyn_tableau_t *tb, /* Tableau to plot. */
    dm_rung_t gopt,       /* Optimal end-rung or {dm_rung_none} */
    char *name,           /* Prefix for file name. */
    char *tag             /* Tag for file name. */
  )
  { /* Create a float-valued image and fill it with {-INF}: */
    dm_image_t *gim;
    int scale;
    dm_alloc_image(nx, ny, &gim, &scale);
    float_image_fill(gim, -INF);

    /* Paint tableau {tb} into float image {gim}, note maximum abs value {maxf}: */
    double maxf = -INF;
    int dr;
    for (dr = 0; dr < tb->nr; dr++)
      { /* Compute coordinnate {r} of rung: */
        int r = tb->minr + dr;
        /* Compute range {mins..maxs} of coord {s} for this {r}: */
        int mins = tb->mins.el[dr];
        int maxs = tb->maxs.el[dr];
        /* Compute number {ns} of distinct coords {s} for this {r}: */
        int ns = (maxs - mins)/2 + 1;
        /* Enumerate {s} coords for this {r}: */
        int ds;
        for (ds = 0; ds < ns; ds++)
          { /* Compute {s} coord: */
            int s = mins + 2*ds;
            /* Compute index of entry in {tb->ev} and get its address {ep}: */
            int k = dr + ds*tb->nr;
            dm_dyn_entry_t *ep = &(tb->ev.el[k]);
            /* Compute the rung {g = (ix,iy)} corresponding to this entry: */ 
            assert ((r + s) % 2 == 0); /* Parity constraint. */
            int ix = (r + s)/2;
            int iy = (r - s)/2;
            /* Reduce indices {ix,iy} modulo {nx,ny}: */ 
            ix = ((ix % nx) + nx) % nx;
            iy = ((iy % ny) + ny) % ny;
            /* Paint score into image (use {max} when reducing): */
            double v = ep->score;
            float *pv = float_image_get_sample_address(gim, 0, ix/scale, iy/scale);
            if (v > (*pv)) { (*pv) = v; }
            /* Update max value: */
            double av = fabs(v);
            if ((av != INF) && (av > maxf)) { maxf = v; }
          }
      }

    /* Convert {gim} to a color pnm image. Maps 0 to gray, {-maxf} to blue, {+maxf} to red. */
    float_image_t *cim = dm_image_colorize_float_image(gim, maxf);
    int NC = cim->sz[0]; /* Channels in color image. */
    assert(NC == 3); /* Hope so. */
    
    /* Paint the optimum path in yellow: */
    float clr[3] = { 0.900, 0.800, 0.000 }; /* Slightly dark reddish yellow */
    dm_rung_t g = gopt;
    while (! dm_rung_is_none(&g))
      { /* Get the indices {ix,iy} from the rung, reduced modulo {nx,ny}: */
        int ix = (g.c[0] % nx + nx) % nx;
        int iy = (g.c[1] % ny + ny) % ny;
        /* Paint the rung into {cim} with yellow color: */
        int c;
        for (c = 0; c < NC; c++) 
          { float_image_set_sample(cim, c, ix/scale, iy/scale, clr[c]); }
        dm_dyn_entry_t *ep = dm_dyn_tableau_get_entry_address(tb, g);
        affirm(ep != NULL, "optmum path falls off the tableau");
        g = ep->prev;
      }
    
    /* Write {cim} to PPM image. */
    dm_write_float_image_as_pnm(cim, 0, 1, name, tag);

    /* Reclaim the image's storage: */
    float_image_free(gim);
    float_image_free(cim);
  }

