#ifndef dm_dyn_H
#define dm_dyn_H

/* Dynamic programming tableaus for incremental optimum pairing. */
/* Last edited on 2006-12-25 23:14:32 by stolfi */

#define dm_dyn_H_COPYRIGHT \
  "Copyright  2005  by the State University of Campinas (UNICAMP)" \
  " and the Federal Fluminense University (UFF)"

#include <dm_basic.h>
#include <dm_rung.h>
#include <dm_pairing.h>

#include <vec.h>
#include <stdint.h>

typedef struct dm_dyn_entry_t
  { double score;    /* Score of best path that ends at {g}. */
    dm_rung_t prev;  /* Previous rung in optimal path to {g}, or {dm_rung_none}. */
  } dm_dyn_entry_t;
  /* Entry of a tableau for dynamic programming, corresponding to some
    rung (index pair) {g}. */

vec_typedef(dm_dyn_entry_vec_t,dm_dyn_entry_vec,dm_dyn_entry_t);
  /* A vector of {dm_dyn_entry_t} variables. */

typedef struct dm_dyn_tableau_t
  { int minr;              /* Minimum R-coord of all entries. */
    int maxr;              /* Maximum R-coord of all entries. */
    int nr;                /* Number of valid R-coord values. */
    int_vec_t mins;        /* The minimum S-coord for each R-coord {r} is {mins[r-minr]}. */
    int_vec_t maxs;        /* The maximum S-coord for each R-coord {r} is {maxs[r-minr]}. */
    int ns;                /* Max number of S-coord values for any {r}. */
    dm_dyn_entry_vec_t ev; /* Tableau entries. */
  } dm_dyn_tableau_t;
  /* A tableau for incremental dynamic programming. Each entry of the
    tableau corresponds to a possible rung {g=(ix,iy)} between two
    sequences, and is stored in {ev[dr + ds*nr]}, where {dr = (ix+iy)
    - minr}, and {ds = ((ix-iy) - mins[dr])/2}. The sum {r = ix+iy} is
    called the /R-coordinate/ of {g}, and the difference {s = ix-iy}
    is called its /S-coordinate/. */
    
dm_dyn_tableau_t dm_dyn_tableau_new(void);
  /* Returns a tableau with zero space, that is, 
    {maxr < minr}, {nr = ns = 0}. */
    
void dm_dyn_tableau_resize(dm_dyn_tableau_t *tb, int minr, int maxr, int maxds);
  /* Resizes the tableau {*tb} for incremental refinement of a given
    pairing. The parameters {minr} and {maxr} must be the minimum and
    maximum R-coordinate {r = ix+iy} of any rung {g = (ix,iy)} that needs to be
    represented in the tableau. The parameter {maxds} must be the maximum
    difference between the S-coordinates {s = ix-iy}
    of any two valid rungs with the same R-coordinate.
    
    The function sets {tb.minr = minr}, {tb.maxr = maxr}, and {tb.ns =
    maxds + 1}. It also recomputes {tb.nr = maxr - minr + 1} and reallocates
    the vectors {tb.mins}, {tb.maxs}, and {tb.ev} as needed to contain
    all elements. The bounds {tb.mins[dr]} and {tb.maxs[dr]} are set
    to an empty interval, for all {dr} in {0..tb.nr-1}. */
    
void dm_dyn_tableau_set_s_range(dm_dyn_tableau_t *tb, int r, int mins, int maxs);
  /* Sets the values of {mins[dr]} and {maxs[dr]} to the given values,
    where {dr = r - minr}. Initializes all tableau elements with R-coordinate
    {r} and S-coordinates in the range {mins..maxs} to have score {-INF} and
    previous rung {dm_rung_none}. Requires {r} in {tb.minr..tb.maxr} and
    {maxs - mins < tb.ns}. */
    
dm_dyn_entry_t *dm_dyn_tableau_get_entry_address(dm_dyn_tableau_t *tb, dm_rung_t g);
  /* Gets the address of the entry of the element of {tb} that
    corresponds to the rung {g}. Assumes that the fields {mins[dr]}
    and {maxs[dr]} have been properly set. Returns NULL if the
    requested element is not in the tableau. */

void dm_dyn_tableau_free(dm_dyn_tableau_t *tb);
  /* Frees the internal storage of tableau {*tb} (but not the 
    descriptor {*tb} itself), and resets it to the empty status. */

void dm_dyn_tableau_image_write_named
  ( int nx,               /* Length of first sequence. */
    int ny,               /* Length of second sequence. */
    dm_dyn_tableau_t *tb, /* Tableau to plot. */
    dm_rung_t gopt,       /* Optimal end-rung or {dm_rung_none} */
    char *name,           /* Prefix for file name. */
    char *tag             /* Tag for file name. */
  );
  /* Writes a PGM image called "{name}{tag}.pgm", with {nx} columns
    and {ny} rows, showing the scores of all rungs spanned by the
    tableau {tb}. The image is downscaled if too big. The sequences
    are assumed to be circular. Eventually it will also plot the optimum
    path that ends with rung {gopt}. */


#endif
