/* See dm_pairing.h */
/* Last edited on 2007-11-01 12:55:21 by stolfi */

#define dm_pairing_C_COPYRIGHT \
  "Copyright  2005  by the State University of Campinas (UNICAMP)" \
  " and the Federal Fluminense University (UFF)"

#include <dm_basic.h>
#include <dm_seq.h>
#include <dm_rung.h>
#include <dm_pairing.h>
#include <dm_image.h>

#include <affirm.h>
#include <sign.h>
#include <fget.h>
#include <nget.h>
#include <jsrandom.h>
#include <jsmath.h>
#include <float_image.h>

#include <assert.h>
#include <stdio.h>
#include <math.h>

/* INTERNAL PAIRING IMPLEMENTATION */

struct dm_pairing_t 
  { int ng;              /* Number of pairs in pairing, or rung index period. */
    int circp;           /* TRUE iff pairing is circular. */
    dm_rung_t gini;      /* Rung number 0. */
    dm_rung_vec_t gv;    /* Rungs of pairing, or empty if perfect. */
  };
  /* The rung count {ng} is always positive.
    
    If {gv.nel} is 0, the pairing is implicit and perfect -- namely,
    {p[i] = gini + (i,i)} for all {i} in {0..ng-1}. If {gv.nel != 0},
    the pairing is explicit; in that case, {gv.nel} must be equal to
    {ng}, {gv[0]} must be equal to {gini}, and {p[i]} is by
    definition {gv[i]} for {i} in {0..ng-1}.

    If {circp} is false, the pairing is open and contains the rungs
    {p[0..ng-1]} only. If {circp} is true, the pairing is circular; in
    that case, the fundamental cycle is {p[0..ng-1]}, and, for all
    {i}, rung {p[i+ng]} is {p[i] + D}, where {D = p[ng] - p[0]}.
    
    If {circp} is true and the pairing is not implicit, then {gv.nel}
    must be {ng+1}, and {p[ng]} is by definition {gv.el[ng]}. */

/* INTERNAL PROTOTYPES */

void dm_pairing_expand(dm_pairing_t *p);
  /* If {*p} is an implicit perfect pairing (with empty rung vector {gv}),
    makes it into an equivalent generic pairing (with non-empty {gv}). */

/* INTERFACE IMPLEMENTATIONS */

vec_typeimpl(dm_rung_vec_t,dm_rung_vec,dm_rung_t);

int dm_pairing_num_rungs(dm_pairing_t *p)
  { return p->ng; }
  
bool_t dm_pairing_is_circular(dm_pairing_t *p)
  { return p->circp; }

dm_rung_t dm_pairing_period(dm_pairing_t *p)
  { int ng = p->ng;
    if (p->circp)
      { if (p->gv.nel == 0)
          { return (dm_rung_t){{ ng, ng }}; }
        else
          { assert(p->gv.nel == ng+1);
            dm_rung_t grep = p->gv.el[ng];
            return (dm_rung_t){{ grep.c[0] - p->gini.c[0],  grep.c[1] - p->gini.c[1] }};
          }
      }
    else
      { return (dm_rung_t){{ 0, 0 }}; }
  }
  
dm_rung_t dm_pairing_get_rung(dm_pairing_t *p, int i)
  { /* Get number of rungs {ng}: */
    int ng = p->ng;
    demand(ng > 0, "invalid rung count");
    /* Get circularity flag {circp}: */
    bool_t circp = dm_pairing_is_circular(p);
    /* Reduce the index {i} to a rung index {k} in the fundamental period: */ 
    int k; 
    if (circp)
      { k = ((i % ng) + ng) % ng; }
    else
      { k = i; demand((k >= 0) && (k < ng), "invalid rung index"); }
    /* Get rung {p[k]}: */
    dm_rung_t g;
    if (p->gv.nel == 0)
      { /* Candidate is perfect and condensed, compute rung: */
        g = p->gini;
        g.c[0] += k; g.c[1] += k;
      }
    else
      { /* Candidate has explicit rungs. */
        assert(p->gv.nel == ng + (circp ? 1 : 0));
        g = p->gv.el[k];
      }
    /* If circular, add period vectors as needed: */
    if (circp)
      { int q = (i - k)/ng; /* Number of periods between {k} and {i}. */
        dm_rung_t per = dm_pairing_period(p);
        g.c[0] += q*per.c[0]; g.c[1] += q*per.c[1]; 
      }
    return g;
  }

bool_t dm_pairing_is_valid (dm_pairing_t *p, bool_t die)
  { int ng = p->ng;
    if (ng <= 0) { fail_test(die, "invalid rung count"); }
    /* Validity of {gv}: */
    if (p->gv.nel < 0) { fail_test(die, "invalid rung vector length"); }
    if (p->gv.nel == 0) 
      { /* Implicit perfect pairing: */
        if (p->gv.el != NULL) { fail_test(die, "non-null vector with zero length"); }
      }
    else
      { /* Explicit pairing: */
        if (p->gv.nel != ng + (p->circp ? 1 : 0)) 
          { fail_test(die, "inconsistent rung vector length"); }
        if (p->gv.el == NULL) 
          { fail_test(die, "null address of non-empty rung vector"); }
        /* First rung must be consistent: */
        if (p->gini.c[0] != p->gv.el[0].c[0]) 
          { fail_test(die, "first rung is inconsistent in X"); }
        if (p->gini.c[1] != p->gv.el[0].c[1]) 
          { fail_test(die, "first rung is inconsistent in X"); }
      }
    return TRUE;
  }
  
dm_pairing_t *dm_pairing_perfect(dm_rung_t *gini, int ng, bool_t circp)
  { /* A pairing must have at least one rung: */
    demand(ng > 0, "invalid rung count for perfect pairing");
    /* Compute final indices: */
    dm_pairing_t *p = (dm_pairing_t *)notnull(malloc(sizeof(dm_pairing_t)), "no mem");
    (*p) = (dm_pairing_t)
      { /*ng*/    ng,
        /*circp*/ circp,
        /*gini*/  *gini,
        /*gv*/    dm_rung_vec_new(0)
      };
    (void)dm_pairing_is_valid(p, TRUE);
    return p;
  }

dm_pairing_t *dm_pairing_from_rung_vec(dm_rung_vec_t *gv, bool_t circp)
  { int ng = gv->nel - (circp ? 1 : 0);
    demand(ng > 0, "invalid rung count");
    dm_pairing_t *p = (dm_pairing_t *)notnull(malloc(sizeof(dm_pairing_t)), "no mem");
    (*p) = (dm_pairing_t)
      { /*ng*/    ng, 
        /*circp*/ circp,
        /*gini*/  gv->el[0],
        /*gv*/    *gv
      };
    (void)dm_pairing_is_valid(p, TRUE);
    return p;
  }

void dm_pairing_expand(dm_pairing_t *p)
  { int ng = p->ng;
    if (p->gv.nel == 0)
      { /* Must be an implicit perfect pairing: */
        assert(ng > 1);
        /* Create a new rung vector and fill it with a perfect pairing: */
        p->gv = dm_rung_vec_new(ng);
        int i;
        dm_rung_t g = p->gini;
        for (i = 0; i < ng; i++)
          { p->gv.el[i] = g; g.c[0]++; g.c[1]++; }
      }
    /* A bit of paranoia can't hurt: */
    (void)dm_pairing_is_valid(p, TRUE);
  }

/* IMPLEMENTATION-INDEPENDENT FUNCTIONS */

bool_t dm_pairing_is_unitary_monotonic(dm_pairing_t *p)
  { int ng = dm_pairing_num_rungs(p);
    demand(ng > 0, "invalid rung count");
    if (p->gv.nel == 0)
      { /* Implicit perfect pairing: */
        return TRUE; /* Always unitary and monotonic. */
      }
    else
      { /* Check steps in {gv}: */
        bool_t circp = dm_pairing_is_circular(p);
        dm_rung_t h = dm_pairing_get_rung(p, 0);
        int k;
        for (k = 1; k < ng + (circp ? 1 : 0); k++)
          { dm_rung_t g = h; 
            /* Get next rung (or first one if circular pairing): */
            h = dm_pairing_get_rung(p, k);
            (void)dm_rung_step_is_unitary_monotonic(&g, &h, TRUE);
          }
      }
    return TRUE;
  }
   
dm_rung_vec_t dm_pairing_collect_rungs(dm_pairing_t *p, int nx, int ny)
  { /* The coordinate periods of a circular pairing must be multiples of {nx,ny}: */
    if (dm_pairing_is_circular(p))
      { dm_rung_t per = dm_pairing_period(p);
        if (nx != 0) { demand(per.c[0] % nx == 0, "X period inconsistent"); }
        if (ny != 0) { demand(per.c[1] % ny == 0, "Y period inconsistent"); }
      }
    /* Get rung count {ngold} of original pairing: */
    int ngold = dm_pairing_num_rungs(p); /* Number of original rungs. */
    /* Reduced rung list is {rgvnew[0..ngnew-1]}: */
    dm_rung_vec_t rgvnew = dm_rung_vec_new(ngold);
    int ngnew = 0; /* Number of distinct reduced rungs. */
    int i;
    for (i = 0; i < ngold; i++)
      { /* Get rung number {i} from pairing: */
        dm_rung_t g = dm_pairing_get_rung(p, i);
        /* Reduce {g} modulo {nx,ny}: */
        if (nx != 0) { g.c[0] = ((g.c[0] % nx) + nx) % nx; }
        if (ny != 0) { g.c[1] = ((g.c[1] % ny) + ny) % ny; }
        /* Find the proper position {j} of rung {g} in the list {rgvnew}: */
        int j = ngnew;
        int cmp;
        while ((j > 0) && ((cmp = dm_rung_lex_compare(&(rgvnew.el[j-1]), &g)) > 0)) { j--; }
        /* Check wether it is a new rung: */
        if ((j == ngnew) || (cmp != 0))
          { /* New rung, insert it in {rgvnew}: */
            dm_rung_vec_expand(&rgvnew, ngnew);
            if (j < ngnew)
              { /* Open up the slot {rgvnew[j]}: */
                int k = ngnew; while (k > j) { rgvnew.el[k] = rgvnew.el[k-1]; k--; }
              }
            rgvnew.el[j] = g;
          }
      }
    return rgvnew;
  }

#define dm_rung_huge (dm_rung_t){{ INT_MAX, INT_MAX }}
  /* A {dm_rung_t} value that should compare higher than any real rung. */

int dm_pairing_count_common_rungs(dm_pairing_t *pa, dm_pairing_t *pb, int nx, int ny)
  { /* Gather the two rung sets, sorted: */
    dm_rung_vec_t gva = dm_pairing_collect_rungs(pa, nx, ny);
    dm_rung_vec_t gvb = dm_pairing_collect_rungs(pb, nx, ny);
    /* Merge the two lists and count coincidences: */
    int nc = 0;
    int ia = 0, ib = 0;
    while ((ia < gva.nel) || (ib < gvb.nel))
      { dm_rung_t ga = (ia < gva.nel ? gva.el[ia] : dm_rung_huge);
        dm_rung_t gb = (ib < gvb.nel ? gvb.el[ib] : dm_rung_huge);
        int cmp = dm_rung_lex_compare(&ga, &gb);
        if (cmp < 0)
          { ia++; }
        else if (cmp > 0)
          { ib++; }
        else
          { nc++; ia++; ib++; }
      }
    /* Cleanup: */
    free(gva.el);
    free(gvb.el);
    return nc;
  }

int64_t dm_pairing_perfect_max_length(int nx, bool_t circx, int ny, bool_t circy)
  { if ((nx == 0) || (ny == 0))
      { return 0; }
    else if (circx && circy)
      { return lcm(nx, ny); }
    else if (circx)
      { return ny; }
    else if (circy)
      { return nx; }
    else
      { return (nx < ny ? nx : ny); }
  }

void dm_pairing_enum_alignments
  ( int nx,       /* Number of bases on first sequence. */
    bool_t circx, /* TRUE if first sequence is circular. */
    int ny,       /* Number of bases on second sequence. */ 
    bool_t circy, /* TRUE if second sequence is circular. */ 
    dm_pairing_use_proc_t *use
  )
  { fprintf(stderr, "x seq: %8d bases circp = %d\n", nx, circx);
    fprintf(stderr, "y seq: %8d bases circp = %d\n", ny, circy);
    bool_t circp = (circx & circy); /* TRUE iff pairings are circular. */
    /* Just in case, check for empty sequences: */
    if ((nx == 0) || (ny == 0)) { return; }
    int idmin, idmax;
    int64_t nrmax = dm_pairing_perfect_max_length(nx, circx, ny, circy); 
    int id;
    /* Must sweep diagonals of the {xp  yp} grid. The diagonals are
      identified by an integer {id} ranging from {idmin} to {idmax}.
      A pair of samples with indices {ix} and {iy} belongs to diagonal
      number {id = ix - iy}. Note that if either sequence is circular
      the diagonals will wrap around. */
    if (circx && circy)
      { idmin = 0; idmax = gcd(nx, ny) - 1; }
    else if (circx)
      { idmin = 0; idmax = +(nx - 1); }
    else if (circy)
      { idmin = -(ny - 1); idmax = 0; }
    else
      { idmin = -(ny - 1); idmax = +(nx + 1); }
    fprintf(stderr, "trying %d alignments\n", idmax - idmin + 1);
    for(id = idmin; id <= idmax; id++)
      { /* fputc(':', stderr);  */
        /* Computes range of pair index {jd} along diagonal: */
        int ix, iy;
        int64_t ng;
        if (circx && circy)
          { ix = +id; iy = 0; ng = nrmax; }
        else if (circx)
          { ix = +id; iy = 0; ng = nrmax; }
        else if (circy)
          { ix = 0; iy = -id; ng = nrmax; }
        else
          { ix = (id < 0 ? 0 : +id); 
            iy = (id > 0 ? 0 : -id);
            ng = nrmax;
            if (ix + ng > nx) { ng = nx - ix; }
            if (iy + ng > ny) { ng = ny - iy; }
          }    
        /* Process the alignment: */
        if (ng > 0) { use(ix, iy, ng, circp); }
      }
    fputc('\n', stderr);
  }

dm_pairing_t *dm_pairing_map_coordinates(dm_pairing_t *p, dm_rung_map_proc_t *map)
  { /* Get rung count {ng} and circularity {circp} of given pairing: */
    int ng = dm_pairing_num_rungs(p);
    bool_t circp = dm_pairing_is_circular(p);
    /* Rungs of new pairing will be {gvnew[0..gvnew.nel-1]}: */
    dm_rung_vec_t gvnew = dm_rung_vec_new(ng + (circp ? 1 : 0));
    /* Map all rungs (including the loopback, if circular): */
    int i;
    for (i = 0; i < gvnew.nel; i++)
      { gvnew.el[i] = map(dm_pairing_get_rung(p, i)); }
    /* Make pairing from rung vector: */
    return dm_pairing_from_rung_vec(&gvnew, p->circp);
  }

dm_pairing_t *dm_pairing_map_to_finer
  ( dm_pairing_t *p, 
    int nxold, 
    int nyold, 
    int nxnew, 
    int nynew, 
    int nwtb
  )
  { 
    auto dm_rung_t map_rung(dm_rung_t g);
    
    dm_rung_t map_rung(dm_rung_t g)
      { int ix = dm_seq_index_map_to_finer(g.c[0], nxold, nxnew, nwtb);
        int iy = dm_seq_index_map_to_finer(g.c[1], nyold, nynew, nwtb);
        return (dm_rung_t){{ix,iy}};
      }
      
    return dm_pairing_map_coordinates(p, &map_rung);
  }

dm_pairing_t *dm_pairing_map_to_coarser
  ( dm_pairing_t *p, 
    int nxold, 
    int nyold, 
    int nxnew, 
    int nynew, 
    int nwtb
  )
  { 
    auto dm_rung_t map_rung(dm_rung_t g);
    
    dm_rung_t map_rung(dm_rung_t g)
      { int ix = dm_seq_index_map_to_coarser(g.c[0], nxold, nxnew, nwtb);
        int iy = dm_seq_index_map_to_coarser(g.c[1], nyold, nynew, nwtb);
        return (dm_rung_t){{ix,iy}};
      }
      
    return dm_pairing_map_coordinates(p, &map_rung);
  }

dm_pairing_t *dm_pairing_interpolate(dm_pairing_t *p)
  { /* Get number of given rungs {ng} and circularity: */
    int ng = dm_pairing_num_rungs(p);
    bool_t circp = dm_pairing_is_circular(p);
    /* Rungs of new pairing will be {gvnew[0..ngnew-1]}: */
    dm_rung_vec_t gvnew = dm_rung_vec_new(ng);
    int ngnew = 0;
    /* Initialize previous rung {g} with first rung: */
    dm_rung_t g = dm_pairing_get_rung(p, 0);
    /* Process all steps of the pairing (including loopback if circular): */
    int i;
    for (i = 1; i < ng + (circp ? 1 : 0); i++)
      { /* Get final rung {h} of step: */
        dm_rung_t h = dm_pairing_get_rung(p, i);
        dm_rung_bresenham(g, h, &ngnew, &gvnew);
      }
    dm_rung_vec_trim(&gvnew, ngnew);
    return dm_pairing_from_rung_vec(&gvnew, p->circp);
  }

dm_pairing_t *dm_pairing_throw
  ( int nx,
    bool_t circx, 
    int ny, 
    bool_t circy, 
    int len,
    bool_t circp, 
    bool_t diagonal,
    double skipProb
  )
  { demand((nx > 0) && (ny > 0), "cannot pair up an empty sequence");
    if (circp) { demand(circx && circy, "a circular pairing requires circular seqs"); }
    dm_pairing_t *p = (dm_pairing_t *)notnull(malloc(sizeof(dm_pairing_t)), "no mem");
    /* Compute the max number of rungs {ngmax}: */
    int ngmax = dm_pairing_perfect_max_length(nx, circx, ny, circy);
    /* Shall we generate a perfect pairing? */
    bool_t perfect = ((skipProb == 0.0) || (drandom() < 0.1));
    if (perfect)
      { /* Generate an implicit perfect pairing. */
        /* Choose the number of rungs {ng}.
          A circular perfect pairing must have precisely {ngmax} rungs.
          An open perfect pairing may have less than that. */
        int ng = (circp ? ngmax : len);
        /* Compute the maximum initial rung {gmax}: */
        dm_rung_t gmin = (dm_rung_t){{ 0, 0 }};
        dm_rung_t gmax = (dm_rung_t){{ nx - (circx ? 1 : ng), ny - (circy ? 1 : ng) }};
        /* Choose the initial rung {gini}: */
        dm_rung_t gini = dm_rung_throw(&gmin, &gmax, diagonal);
        /* Generate pairing: */
        p = dm_pairing_perfect(&gini, ng, circp);
      }
    else
      { /* Generate an explicit pairing. */
        /* Choose the initial rung {gini} and final (or loopback) rung {gfin}. */
        dm_rung_t gini; /* Initial rung. */
        dm_rung_t gfin; /* Final rung (loopback if {circp}). */
        if (circp)
          { dm_rung_t gmin = (dm_rung_t){{ 0, 0 }};
            dm_rung_t gmax = (dm_rung_t){{ nx-1, ny-1 }};
            gini = dm_rung_throw(&gmin, &gmax, diagonal);
            /* The coord periods must be multiples of the seq lengths.
              In most applications it makes no sense to have pairings 
              that make funny numbers of turns around each seq.
              So we generate a unitary monotonic pairing that does 
              {ny} turns around the X-sequence and {nx} turns around
              the Y-sequence:
            */
            int DXY = lcm(nx, ny);
            gfin = (dm_rung_t){{ gini.c[0] + DXY, gini.c[1] + DXY }};
          }
        else
          { /* The pairing is open. */
            /* Coose a segment of the X sequence (possibly with wrap-around): */
            int xini, xfin;
            dm_seq_throw_segment(nx, circx, len, &xini, &xfin);
            /* Coose a segment of the Y sequence (possibly with wrap-around): */
            int yini, yfin;
            dm_seq_throw_segment(ny, circy, len, &yini, &yfin);
            /* The pairing should span both segments: */
            gini = (dm_rung_t){{ xini, yini }};
            gfin = (dm_rung_t){{ xfin, yfin }};
          }
        /* Now generate a random unitary monotonic path from {gini} to {gfin}: */
        dm_rung_vec_t gv = dm_rung_vec_throw(&gini, &gfin, skipProb);
        /* Wrap it up as a pairing: */
        p = dm_pairing_from_rung_vec(&gv, circp);
      }
    return p;
  }

void dm_pairing_write(FILE *wr, int nx, int ny, dm_pairing_t *p, int ixsize)
  {
    /* Write number of pairs {p->ng}: */
    int ng = p->ng;
    demand(ng > 0, "invalid {ng}");
    fprintf(wr, "%*d", ixsize, ng);
    
    /* Write circularity flag: */
    fprintf(wr, " %c", (p->circp ? 'T' : 'F'));

    /* Write initial rung: */
    fprintf(wr, " %*d %*d", ixsize, p->gini.c[0], ixsize, p->gini.c[1]);

    /* Print rung vector and check : */
    if (p->gv.nel == 0)
      { /* Perfect pairing with implicit rung vector: */ 
        fprintf(wr, " *");
      }
    else
      { /* Pairing with explicit rungs: */ 
        demand(p->gv.nel == ng + (p->circp ? 1 : 0), "inconsistent num of pairs");
        /* Print code of first pair (always "|"). */
        fprintf(wr, " |");
        /* Print codes of other pairs: */
        dm_rung_t h = p->gini; /* Last printed rung. */
        int i; 
        for (i = 1; i < p->gv.nel; i++)
          { dm_rung_t g = h; 
            h = dm_pairing_get_rung(p, i);
            dm_rung_step_write(wr, &g, &h);
          }
      }
    fflush(wr);
  }

dm_pairing_t *dm_pairing_read(FILE *rd)
  {
    /* Number of rungs: */
    int ng = fget_int(rd);
    demand(ng > 0, "bad number of pairs");
    
    /* Circularity flag: */
    bool_t circp = fget_bool(rd);
    
    /* Initial rung: */
    dm_rung_t gini;
    gini.c[0] = fget_int(rd); demand(gini.c[0] >= 0, "bad init index for {x} sequence"); 
    gini.c[1] = fget_int(rd); demand(gini.c[1] >= 0, "bad init index for {y} sequence"); 
    
    /* Read first char of pairing: */
    fget_skip_spaces(rd); 
    int ch = fget_char(rd);
    
    if (ch == '*')
      { /* Perfect matching: */
        return dm_pairing_perfect(&gini, ng, circp);
      }
    else if (ch == '|')
      { /* Generic pairing. */
        /* Allocate the rung vector: */
        int gvsize = ng + (circp ? 1 : 0);
        dm_rung_vec_t gv = dm_rung_vec_new(gvsize);
        /* Save initial pair: */
        dm_rung_t g = gini;
        gv.el[0] = g;
        /* Read and save remaining pairs: */
        int i; 
        for(i = 1; i < gv.nel; i++)
          { dm_rung_t h = dm_rung_step_read(rd, g); 
            gv.el[i] = h;
            g = h;
          }
        return dm_pairing_from_rung_vec(&gv, circp);
      }
    else
      { demand(FALSE, "bad or missing initial pairing character");
        return NULL; /* Compiler pacifier. */
      }
  }

void dm_pairing_paint(dm_pairing_t *p, int nx, int ny, float v, float_image_t *img, int scale)
  { /* Get number of rungs in candidate's pairing: */
    int ng = dm_pairing_num_rungs(p);
    /* Enumerate sample pairs in candidate: */
    int ir;
    for (ir = 0; ir < ng; ir++)
      { dm_rung_t g = dm_pairing_get_rung(p, ir);
        /* Get indices into sequences: */
        int32_t ix, iy;
        ix = g.c[0]; iy = g.c[1];
        /* Reduce indices modulo {nx,ny}: */
        ix = ((ix % nx) + nx) % nx;
        iy = ((iy % ny) + ny) % ny;
        /* Add average rung score to pixel: */
        dm_bump_image(img, ix, iy, scale, v);
      }
  }
  
void dm_pairing_seq_seq_image_write_named
  ( dm_seq_t *xp, 
    dm_seq_t *yp, 
    dm_step_score_proc_t *step_score, 
    char *name, 
    char *tag
  )
  { int nx = xp->dv.nel; bool_t circx = dm_seq_is_circular(xp);
    int ny = yp->dv.nel; bool_t circy = dm_seq_is_circular(yp);
    dm_image_t *gim = NULL;
    int scale = -1;

    dm_alloc_image(nx, ny, &gim, &scale);
    fprintf(stderr, "  image size = %d  %d  scale factor = %d\n", nx, ny, scale);
    
    /* Clear all totals: */
    float_image_fill(gim, 0.0);
     
    /* Count coincidence points: */
    int iy, ix;
    for (iy = 0; iy < ny; iy++)
      { for (ix = 0; ix < nx; ix++)
          { dm_rung_t g = (dm_rung_t){{ ix, iy }};
            double score = 0;
            double maxssc = -INF; /* Maximum step score into {g}. */

            auto void try_step(dm_rung_t f);
              /* Tries the step from {f} to {g}. */

            void try_step(dm_rung_t f)
              { if ((f.c[0] < 0) && (!circx)) { return; }
                if ((f.c[1] < 0) && (!circy)) { return; }
                double ssc = step_score(f, g, xp, yp);
                if (ssc > maxssc) { maxssc = ssc; }
              }

            /* Evaluate the three steps into {g}: */
            try_step((dm_rung_t){{ ix-1, iy-1 }});
            try_step((dm_rung_t){{ ix,   iy-1 }});
            try_step((dm_rung_t){{ ix-1, iy   }});
            
            /* Evaluate the score for starting at {g}: */
            /* try_step(dm_rung_none); */

            score = maxssc;
            dm_bump_image(gim, ix, iy, scale, (float)score);
          }
      }
    
    /* Convert to PPM image: */
    double lov, hiv;
    dm_compute_min_max_image_range(gim, &lov, &hiv);
    double maxf = (fabs(lov) > fabs(hiv) ? fabs(lov) : fabs(hiv));
    
    /* Colorize image: */
    float_image_t *cim = dm_image_colorize_float_image(gim, maxf);
    
    /* Write image: */
    dm_write_float_image_as_pnm(cim, 0, 1, name, tag);
    
    /* Reclaim image space: */
    float_image_free(gim);
    float_image_free(cim);
  }
  
double dm_pairing_perfect_score
  ( dm_seq_t *xp,
    int ix, 
    dm_seq_t *yp,
    int iy, 
    int ng, 
    bool_t circp,
    dm_step_score_proc_t *step_score   /* Step scoring function. */
  )
  { int nx = xp->dv.nel;
    int ny = yp->dv.nel;
    ix %= nx;
    iy %= ny;
    if (ng == 0) { return 0.0; }
    int64_t t;
    dm_rung_t g = (dm_rung_t){{ix, iy}};
    double sc = (circp ? 0 : step_score(dm_rung_none, g, xp, yp));
    for (t = 1; t < ng; t++)
      { dm_rung_t h = (dm_rung_t){{ix + t, iy + t}};
        sc += step_score(g, h, xp, yp);
        g = h;
      }
    /* Final step: */
    dm_rung_t h = (circp ? (dm_rung_t){{ix + ng, iy + ng}} : dm_rung_none);
    sc += step_score(g, h, xp, yp);
    return sc;
  }

