/* See dm_psplot.h */
/* Last edited on 2006-12-30 21:11:12 by stolfi */

#define dm_psplot_C_COPYRIGHT \
  "Copyright  2006  by the State University of Campinas (UNICAMP)"

#include <dm_basic.h>
#include <dm_psplot.h>

#include <pswr.h>
#include <affirm.h>

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>

/* INTERNAL REPRESENTATION */

struct dm_psplot_t
  { PSStream *ps;
    /* Usable figure dimensions in mm */
    double size[2]; 
    /* Client-to-PSStream coordinate conversion: */
    double cMin[2], zMin[2];
    double cMax[2], zMax[2];
  };
  /* The Postscript stream {ps} is set up so that the plot area is
    {size[0]} by {size[1]} mm, and the {pswr.h} client coordinates
    (here called /device coordinates/) are measured in mm from the
    lower left corner.
    
    The {dm_psplot.h} client coordinates are such that the rectangle
    {[zMin[0] _ zMax[0]]  [zMin[1] _ zMax[1]]} in client coords gets
    plotted to the rectangle {[cMin[0]_cMax[0]]  [cMin[1]_cMax[1]]}
    in device coords. */

dm_psplot_t *dm_psplot_new_named
  ( char *name,
    char *tag,
    double hSize, 
    double vSize,
    double mrg
  )
  { /* Join {name} and {tag} into a file name prefix: */
    char *prefix = NULL;
    asprintf(&prefix, "%s%s", name, tag);
    /* Add the extension ".eps" and open the file: */
    FILE *wr = dm_open_write(name, tag, ".eps", TRUE);
    /* Create the {dm_psplot_t} object stream from it: */
    dm_psplot_t *dp = dm_psplot_new(wr, prefix, hSize, vSize, mrg);
    /* Cannot free the {prefix} here --- it's in use! */
    return dp;
  }

dm_psplot_t *dm_psplot_new
  ( FILE *wr,
    char *prefix,
    double hSize, 
    double vSize,
    double mrg
  )
  { /* Create an Encapsulated Postscript stream from {wr}: */
    double mm = dm_PT_PER_MM;
    double hPageSize = hSize + 2*mrg;
    double vPageSize = vSize + 2*mrg;
    PSStream *ps = pswr_new_stream(TRUE, prefix, wr, NULL, hPageSize*mm, vPageSize*mm);
    /* Sets the page layout to a single picture {hSize} by {vSize}: */
    pswr_set_page_layout(ps, hSize*mm, vSize*mm, FALSE, mrg*mm, mrg*mm, 0, 1, 1);
    /* Sets the client plot window to {[0 _ hSize]  [0 _ vSize]}, in mm: */
    pswr_new_picture(ps, 0, hSize, 0, vSize); 
    /* Create the {dm_psplot_t} object: */
    dm_psplot_t *dp = (dm_psplot_t *)notnull(malloc(sizeof(dm_psplot_t)), "no mem");
    dp->ps = ps;
    dp->size[0] = hSize; dp->size[1] = vSize;
    /* Set default client and device reference window: */
    int ax;
    for (ax = 0; ax < 2; ax++)
      { /* Device ref window is the whole plot area: */
        dp->cMin[ax] = 0; dp->cMax[ax] = dp->size[ax];
        /* Client ref window is set so that cient and device coords are the same: */
        dp->zMin[ax] = dp->cMin[ax]; dp->zMax[ax] = dp->cMax[ax];
      }
    return dp;
  }

dm_psplot_t *dm_psplot_new_graph
  ( FILE *wr,
    char *prefix,
    double hGraphSize, 
    double vGraphSize,
    bool_t scaleL, bool_t titleL,
    bool_t scaleR, bool_t titleR,
    bool_t scaleB, bool_t titleB,
    bool_t scaleT, bool_t titleT,
    double fontSize,
    int maxLabChars,
    double mrg
  )
  {
    double mm = dm_PT_PER_MM;
    
    /* Estimate the character dimensions in mm: */
    double vCharSize = fontSize/mm; /* Estimated character height, with skosh. */
    double hCharSize = 0.8*vCharSize; /* Reasonable guess? */
    
    /* Compute the max label dimensions in mm: */
    double hLabelSize = maxLabChars*hCharSize;
    double vLabelSize = vCharSize;
    
    /* Initialize title/scale margin widths in mm: */
    double mrgL = 0.0; /* Left. */
    double mrgR = 0.0; /* Right. */
    double mrgB = 0.0; /* Bottom. */
    double mrgT = 0.0; /* Top. */
    
    /* Add space for X and Y titles. Assume that Y titles are rotated. */
    if (titleL) { mrgL += vCharSize; }
    if (titleR) { mrgR += vCharSize; }
    if (titleB) { mrgB += vCharSize; }
    if (titleT) { mrgT += vCharSize; }
    
    /* Add space for X and Y scales. Assume that Y scales are NOT rotated. */
    if (scaleL) { mrgL += hLabelSize; }
    if (scaleR) { mrgR += hLabelSize; }
    if (scaleB) { mrgB += vLabelSize; }
    if (scaleT) { mrgT += vLabelSize; }
    
    /* Make sure that there is enough space for overshoot of scales: */
    double hOver = 0.50*hLabelSize; /* Estimated H overshoot of X scale labels */
    double vOver = 0.75*vLabelSize; /* Estimated V overshoot of Y scale labels. */
    if (scaleL || scaleR)
      { if (mrgB < vOver) { mrgB = vOver; }
        if (mrgT < vOver) { mrgT = vOver; }
      }
    if (scaleB || scaleT)
      { if (mrgL < hOver) { mrgL = hOver; }
        if (mrgR < hOver) { mrgR = hOver; }
      }
    
    /* Compute dimensions of plottable area (graph,titles,scales): */
    double hTotSize = hGraphSize + mrgL + mrgR;
    double vTotSize = vGraphSize + mrgB + mrgT;

    /* Create plot object: */
    dm_psplot_t *dp = dm_psplot_new(wr, prefix, hTotSize, vTotSize, mrg);
    
    /* Set reference rectangles to graph area only: */
    double hMin = 0 + mrgL, hMax = hTotSize - mrgR;
    double vMin = 0 + mrgB, vMax = vTotSize - mrgT;
    dm_psplot_set_device_ref_window(dp, hMin, hMax, vMin, vMax);
    dm_psplot_set_client_ref_window(dp, hMin, hMax, vMin, vMax);
    
    return dp;
  }

PSStream *dm_psplot_get_ps_stream(dm_psplot_t *dp)
  { return dp->ps; }

void dm_psplot_get_plot_size(dm_psplot_t *dp, double *hSize, double *vSize)
  { *hSize = dp->size[0]; *vSize = dp->size[1]; }
   
void dm_psplot_set_device_ref_window
  ( dm_psplot_t *dp,
    double hMin, double hMax, 
    double vMin, double vMax
  )
  { dp->cMin[0] = hMin; dp->cMax[0] = hMax;
    dp->cMin[1] = vMin; dp->cMax[1] = vMax;
  }

void dm_psplot_set_client_ref_window
  ( dm_psplot_t *dp,
    double xMin, double xMax, 
    double yMin, double yMax
  )
  { dp->zMin[0] = xMin; dp->zMax[0] = xMax;
    dp->zMin[1] = yMin; dp->zMax[1] = yMax;
  }

void dm_psplot_shrink_device_ref_window
  ( dm_psplot_t *dp,
    double lMrg, double rMrg, 
    double bMrg, double tMrg
  )
  { dp->cMin[0] += lMrg; dp-> cMax[0] -= rMrg;
    dp->cMin[1] += bMrg; dp-> cMax[1] -= tMrg;
  }

void dm_psplot_expand_client_ref_window
  ( dm_psplot_t *dp,
    double lMrg, double rMrg, 
    double bMrg, double tMrg
  )
  { dp->zMin[0] -= lMrg; dp-> zMax[0] += rMrg;
    dp->zMin[1] -= bMrg; dp-> zMax[1] += tMrg;
  }

double dm_psplot_map_x(dm_psplot_t *dp, double x)
  { double s = (x - dp->zMin[0])/(dp->zMax[0] - dp->zMin[0]), r = 1-s;
    return r*dp->cMin[0] + s*dp->cMax[0];
  }
  
double dm_psplot_map_y(dm_psplot_t *dp, double y)
  { double s = (y - dp->zMin[1])/(dp->zMax[1] - dp->zMin[1]), r = 1-s;
    return r*dp->cMin[1] + s*dp->cMax[1];
  }

double dm_psplot_map_coord(dm_psplot_t *dp, PSAxis axis, double coord)
  { demand((axis == HOR) || (axis == VER), "bad axis");
    double s = (coord - dp->zMin[axis])/(dp->zMax[axis] - dp->zMin[axis]), r = 1-s;
    return r*dp->cMin[axis] + s*dp->cMax[axis];
  }
  
void dm_psplot_map_coords(dm_psplot_t *dp, double x, double y, double *h, double *v)
  { (*h) = dm_psplot_map_x(dp, x);
    (*v) = dm_psplot_map_y(dp, y);
  }

/* DEVICE -> CLIENT */

double dm_psplot_unmap_h(dm_psplot_t *dp, double h)
  { double s = (h - dp->cMin[0])/(dp->cMax[0] - dp->cMin[0]), r = 1-s;
    return r*dp->zMin[0] + s*dp->zMax[0];
  }
  
double dm_psplot_unmap_v(dm_psplot_t *dp, double v)
  { double s = (v - dp->cMin[1])/(dp->cMax[1] - dp->cMin[1]), r = 1-s;
    return r*dp->zMin[1] + s*dp->zMax[1];
  }

double dm_psplot_unmap_coord(dm_psplot_t *dp, PSAxis axis, double coord)
  { demand((axis == HOR) || (axis == VER), "bad axis");
    double s = (coord - dp->cMin[axis])/(dp->cMax[axis] - dp->cMin[axis]), r = 1-s;
    return r*dp->zMin[axis] + s*dp->zMax[axis];
  }
  
void dm_psplot_unmap_coords(dm_psplot_t *dp, double h, double v, double *x, double *y)
  { (*x) = dm_psplot_unmap_h(dp, h);
    (*y) = dm_psplot_unmap_v(dp, v);
  }

/* AXES, TICKS, ETC */

void dm_psplot_draw_segment(dm_psplot_t *dp, double xa, double ya, double xb, double yb)
  { /* Get the device coordinate of the points: */
    double ha, va;
    dm_psplot_map_coords(dp, xa, ya, &ha, &va);
    double hb, vb;
    dm_psplot_map_coords(dp, xb, yb, &hb, &vb);
    /* Plot the segment: */
    pswr_segment(dp->ps, ha, va, hb, vb);
  }

void dm_psplot_draw_ref_axis(dm_psplot_t *dp, PSAxis axis)
  { pswr_set_pen(dp->ps, 0.25,0.25,0.25, 0.10, 0.0, 0.0);
    /* Get the device coordinate of the requested axis along the other axis: */
    double cpos = dm_psplot_map_coord(dp, 1-axis, 0.0);
    pswr_axis(dp->ps, axis, cpos, dp->cMin[axis], dp->cMax[axis]);
  }

void dm_psplot_draw_ref_frame(dm_psplot_t *dp)
  { pswr_set_pen(dp->ps, 0,0,0, 0.10, 0.0, 0.0);
    pswr_rectangle(dp->ps, dp->cMin[0], dp->cMax[0], dp->cMin[1], dp->cMax[1], FALSE, TRUE);
  }

void dm_psplot_draw_plot_frame(dm_psplot_t *dp)
  { pswr_set_pen(dp->ps, 0,0,0, 0.10, 0.0, 0.0);
    pswr_rectangle(dp->ps, 0, dp->size[0], 0, dp->size[1], FALSE, TRUE);
  }
  
void dm_psplot_draw_tick
  ( dm_psplot_t *dp, 
    PSAxis axis, 
    double xt, 
    double yt,
    double tickSize,
    double tickAlign,
    char *label,
    double labAlign
  )
  { /* Compute nominal device coordinates {ht,vt} of tick: */
    double ht = dm_psplot_map_x(dp, xt);
    double vt = dm_psplot_map_y(dp, yt);
    /* Plot the tick proper: */
    pswr_tick(dp->ps, axis, ht, vt, tickSize, tickAlign);
    if (label != NULL)
      { /* Decide the vertical and horizontal alignment of label: */
        double hAlign = (axis == HOR ? 0.5 : labAlign);
        double vAlign = (axis == VER ? 0.5 : labAlign);
        /* Adjust the position of the label's reference point to account for tick size: */
        double cadj = (1-labAlign-tickAlign)*tickSize; 
        double hb = (axis == HOR ? ht : ht + cadj);
        double vb = (axis == VER ? vt : vt + cadj);
        pswr_label(dp->ps, label, hb, vb, hAlign, vAlign);
      }
  }

void dm_psplot_draw_scale
  ( dm_psplot_t *dp, 
    PSAxis axis, 
    double pos,
    double tickSize,
    double tickAlign,
    double tickMinDist, 
    double tickMinStep, 
    char *fmt,
    double labAlign,
    double labMinDist,
    double labMinStep
  )
  { 
    /* Oh paranoia, not even Goya could draw'ya: */
    demand((axis == HOR) || (axis == VER), "bad axis");
    
    /* Get the raw span of ticks in device coords (mm): */
    double cMin = dp->cMin[axis];
    double cMax = dp->cMax[axis];
    
    /* Swap if needed to ensure {cMin <= cMax}: */
    if (cMin > cMax) { double c = cMin; cMin = cMax; cMax = c; }
    
    /* Leave some space {eps} between range limits and first/last ticks: */
    double eps = 0.5; /* In mm. */
    cMin += eps; cMax -= eps;
    
    /* Choose the client coords of minor ticks (0 if no ticks in range): */
    double ztMin, ztMax; /* Min and max client coords of minor ticks. */
    double ztStep; /* Client coord increment between minor ticks (0 if no ticks). */
    dm_psplot_choose_tick_coords
      ( dp, axis, cMin, cMax, tickMinDist, tickMinStep, &ztMin, &ztMax, &ztStep );
    /* fprintf(stderr, "  ztMin = %24.16e ztMax = %24.16e ztStep = %24.16e\n", ztMin, ztMax, ztStep); */
    if (ztStep <= 0) { /* Not enough space for ticks: */ return; }
    
    /* Compute number {ntsteps} of minor tick steps: */
    int ntsteps = (int)floor((ztMax/ztStep) - (ztMin/ztStep) + 0.5);
    assert(ntsteps >= 0); 
    
    /* Compute minor-to-major tick ratio {labPer} (0 if no major ticks) and phase {labSkp}: */
    int labPer;   /* Minor tick steps in one major tick step. */
    int labSkp; /* Minor ticks before the first major tick. */
    if (fmt == NULL)
      { /* Clients wants no major ticks: */
        labPer = labSkp = 0;
      }
    else 
      { /* Choose major ticks honoring {labSp} and {minStep}: */
        dm_choose_label_coords
          ( dp, axis, ztMin, ztMax, ztStep, labMinDist, labMinStep, &labPer, &labSkp);
      }
      
    /* Set pen to solid thin black: */
    pswr_set_pen(dp->ps, 0,0,0, 0.10, 0.0, 0.0);
    
    int i;
    for (i = 0; i <= ntsteps; i++)
      { /* Get fractions {r:s} of {i} betwen 0 and {ntsteps}: */
        double s = (ntsteps == 0 ? 0.5 : ((double)i)/((double)ntsteps)), r = 1-s;
        /* Compute nominal client coordinate {zt} of tick on the given {axis}: */
        double zt = r*ztMin + s*ztMax;
        /* Nominal client coords of tick mark are {xt,yt}: */
        double xt = (axis == HOR ? zt : pos);
        double yt = (axis == VER ? zt : pos);
        /* Decide which kind of label to draw: */
        if ((labPer > 0) && (i % labPer == labSkp))
          { /* Draw a major (labeled) tick: */
            assert(fmt != NULL);
            /* Format the text: */
            char *label = NULL;
            asprintf(&label, fmt, zt);
            /* Draw a labeled tick: */
            dm_psplot_draw_tick(dp, axis, xt, yt, 1.5*tickSize, tickAlign, label, labAlign);
            free(label);
          }
        else
          { /* Draw a minor (label-less) tick: */
            dm_psplot_draw_tick(dp, axis, xt, yt, tickSize, tickAlign, NULL, 0.0);
          }
      }
  }

void dm_psplot_choose_tick_coords
  ( dm_psplot_t *dp, 
    PSAxis axis, 
    double cMin, 
    double cMax, 
    double minDist, 
    double minStep, 
    double *zMinP,
    double *zMaxP,
    double *zStepP
  )
  { /* Check required conditions: */
    demand((minDist > 0) || (minStep > 0), "neither minDist nor MinStep were given"); 
    /* Local copies of results: */
    double zMin, zMax, zStep;
    /* Any space at all? */
    if (cMin > cMax) 
      { /* Interval is empty, no ticks there: */
        zMin = +INF; zMax = -INF; zStep = 0;
      }
    else
      { /* Compute the client increment {zStep} equivalent to device increment {labSp}: */
        double zDif = fabs(dp->zMax[axis] - dp->zMin[axis]);
        double cDif = fabs(dp->cMax[axis] - dp->cMin[axis]);
        zStep = minDist*(zDif/cDif);
        if (zStep < minStep) { zStep = minStep; }
        assert(zStep >= 0);
        /* Round {zStep} to a nice value: */
        zStep = pswr_round_to_nice(zStep);
        /* Get range {[zMin_zMax]} of client coords corresponding to {[cMin_cMax]}: */
        zMin = dm_psplot_unmap_coord(dp, axis, cMin);
        zMax = dm_psplot_unmap_coord(dp, axis, cMax);
        if (zMin > zMax) { double z = zMin; zMin = zMax; zMax = z; }
        /* fprintf(stderr, "  zMin =  %24.16e zMax =  %24.16e  zStep = %24.16e\n", zMin, zMax, zStep); */
        /* See whether there are any multiples of {zStep} in {zMin,zMax}: */
        assert(zMin <= zMax);
        double qMin = ceil(zMin/zStep);
        double qMax = floor(zMax/zStep);
        /* fprintf(stderr, "  qMin =  %24.16e qMax =  %24.16e\n", qMin, qMax); */
        if (qMin <= qMax)
          { /* Synchronize {zMin,zMax} to multiples of {zStep}: */
            double zrMin = qMin*zStep; assert(zrMin >= zMin); 
            double zrMax = qMax*zStep; assert(zrMax <= zMax);
            /* fprintf(stderr, "  zrMin = %24.16e zrMax = %24.16e\n", zrMin, zrMax); */
            assert(zrMin <= zrMax);
            zMin = zrMin; zMax = zrMax;
          }
        else
          { /* No multiples of {zStep} in range: */
            zMin = +INF; zMax = -INF; zStep = 0;
          }
      }
    /* fprintf(stderr, "  zMin =  %24.16e zMax =  %24.16e  zStep = %24.16e\n", zMin, zMax, zStep); */
    (*zMinP) = zMin; (*zMaxP) = zMax; (*zStepP) = zStep;
  }

void dm_choose_label_coords
  ( dm_psplot_t *dp,
    PSAxis axis,
    double ztMin,
    double ztMax, 
    double ztStep,
    double minDist,
    double minStep,
    int *labPerP,
    int *labSkpP
  )
  {
    double eps = 1.0e-12; /* Relative fudge factor to compensate for roundoff errors. */

     /* First, we choose the label period {labPer}: */
    int labPer; 
    
    auto int min_mult(int k, double unit, double min);
      /* Increments the positive integer {k} so that {k*unit >= minv},
        allowing for some roundoff noise in {unit}. */
         
    int min_mult(int k, double unit, double minv)
      { unit = (1+eps)*unit;
        if (k*unit < minv)
          { k = (int)ceil(minv/unit);
            assert(k*unit >= minv);
          }
        return k;
      }
      
    if ((ztStep == 0) || (ztMin > ztMax))
      { /* No ticks, no major ticks: */
        labPer = 0;
      }
    else
      { /* Set {labPer} to the minmum period that satisfies {labSp}: */
        double tickDist; /* Actual tick spacing {tickDist} in device coords (fudged up): */
        tickDist = fabs(dm_psplot_map_coord(dp, axis, ztStep) - dm_psplot_map_coord(dp, axis, 0));
        labPer = min_mult(1, tickDist, minDist);
        assert(labPer > 0);

        /* Adjust {labPer} upwards to honor {minStep}: */
        demand(ztStep > 0, "invalid ztStep"); 
        labPer = min_mult(labPer, ztStep, minStep);
        assert(labPer > 0);

        /* Adjust {labPer} so that the label value increment  is nice: */
        if (labPer > 1)
          { /* Compute the label step {zbStep} in client coords: */
            double zbStep = labPer*ztStep;
            /* Increment {labPer} until {zbStep} is nice (should't be too far): */
            while (TRUE)
              { assert(zbStep > minStep);
                double zbNice = pswr_round_to_nice(zbStep);
                /* fprintf(stderr, "    labPer = %d zbStep =  %24.16e", labPer, zbStep); */
                /* fprintf(stderr, " zbNice = %24.16e\n", zbNice); */
                if (zbNice == INF) { /* Overflow, give up: */ labPer = 0; break; }
                if (zbNice <= zbStep) { /* Phew! */ break; }
                int labPerNew = min_mult(labPer, ztStep, zbNice);
                assert(labPerNew > labPer);
                labPer = labPerNew;
                zbStep = labPer*ztStep;
              }
          }
      }
      
    /* Compute the number {labSkp} of minor ticks before first major tick: */
    int labSkp;
    if (labPer == 0)
      { /* Just in case: */ labSkp = 0; }
    else
      { /* Compute coords {zbMin,zbMax} of first and last major ticks: */
        double zbStep = labPer*ztStep; /* Label spacing in client coords. */
        double qbMin = ceil(ztMin/zbStep);
        double qbMax = floor(ztMax/zbStep);
        /* fprintf(stderr, "  qbMin =  %24.16e qbMax =  %24.16e\n", qbMin, qbMax); */
        if (qbMin <= qbMax)
          { /* There is a multiple of {zbStep} in the range: */
            double zbMin = qbMin*zbStep;
            labSkp = (int)floor((zbMin - ztMin)/ztStep + 0.5);
            /* fprintf(stderr, "  zbMin  =  %24.16e labSkp = %d\n", zbMin, labSkp); */
            assert((labSkp >= 0) && (labSkp < labPer));
          }
        else
          { /* No multiples of {zbStep} in range: */
            labPer = labSkp = 0;
          }
      }
    
    /* Return results: */
    (*labPerP) = labPer; (*labSkpP) = labSkp;
  }

void dm_psplot_draw_y_polyline
  ( dm_psplot_t *dp,
    double xMin, 
    double xMax,
    double y[],
    int n
  )
  { if (n < 2) { return; }
    double n1 = n - 1;
    double h0, v0; /* Previously plotted point. */
    dm_psplot_map_coords(dp, xMin, y[0], &h0, &v0);
    int i;
    for (i = 1; i < n; i++)
      { /* Plot step from point {i-1} to point {i}: */
        double s = i/n1, r = 1-s;
        double x1 = r*xMin + s*xMax;
        double h1, v1;
        dm_psplot_map_coords(dp, x1, y[i], &h1, &v1);
        pswr_segment(dp->ps, h0, v0, h1, v1);
        h0 = h1; v0 = v1;
      }
  }

void dm_psplot_draw_y_dots
  ( dm_psplot_t *dp,
    double xMin, 
    double xMax,
    double y[],
    int n,
    double rad,
    bool_t fill,
    bool_t draw
  )
  { if (n < 1) { return; }
    double n1 = n - 1;
    int i;
    for (i = 0; i < n; i++)
      { /* Plot point {i}: */
        double s = i/n1, r = 1-s;
        double x = r*xMin + s*xMax;
        double h, v;
        dm_psplot_map_coords(dp, x, y[i], &h, &v);
        pswr_dot(dp->ps, h, v, rad, fill, draw);
      }
  }

void dm_psplot_close(dm_psplot_t *dp)
  { pswr_close_stream(dp->ps);
    free(dp);
  }
