#ifndef dm_rung_H
#define dm_rung_H

/* A rung is a pair of datum indices, one from each sequence. */
/* Last edited on 2007-01-04 01:53:18 by stolfi */

#define dm_rung_H_COPYRIGHT \
  "Copyright  2006  by the State University of Campinas (UNICAMP)"

#include <dm_basic.h>
#include <dm_seq.h>

#include <sign.h>
#include <vec.h>
#include <float_image.h>

#include <stdint.h>

/* RUNGS
  
  A `rung' is a pair of integers. It can be seen as a point of the
  integer plane, so its two components are called the `X coordinate'
  and `Y coordinates', or `0-coordinate' and `1-coordinate'.
  
  (Each coordinate of a rung is usually meant to be an index into some
  sequence, but this interpretation is not used to this interface,
  exacept as a motivation). */

typedef struct dm_rung_t { int32_t c[2]; } dm_rung_t; 

#define dm_rung_X(g) ((g).c[0])
#define dm_rung_Y(g) ((g).c[1])
  /* The X- and Y-coordinates of a rung. */
  
#define dm_rung_R(g) ((g).c[0]+(g).c[1])
#define dm_rung_S(g) ((g).c[0]-(g).c[1])
  /* The R- and S-coordinates of a rung (along the main
    diagonal and across it, respectively). */

#define dm_rung_none (dm_rung_t){{ INT_MAX, INT_MAX }}
  /* A special {dm_rung_t} value mening `no rung'. */

bool_t dm_rung_is_none(dm_rung_t *g);
  /* TRUE iff {g == dm_rung_none}. */

bool_t dm_rung_equal(dm_rung_t *ga, dm_rung_t *gb);
  /* TRUE iff {ga == gb}. */

sign_t dm_rung_lex_compare(dm_rung_t *ga, dm_rung_t *gb);
  /* Compares two rungs {ga,gb} by lexicograpic order. Namely, compares
    the X coordinates, and breaks ties comparing the Y coordinates.
    The result is {NEG} (-1) if {ga} precedes {gb}, {POS} (+1) 
    if {ga} follows {gb}, and {ZER} (0) if {ga == gb}. */
  
sign_t dm_rung_strict_compare(dm_rung_t *ga, dm_rung_t *gb);
  /* Compares two rungs {ga,gb} by strict domination order. The result
    is {NEG} (-1) if {dm_rung_X(ga) <= dm_rung_X(gb)} and
    {dm_rung_Y(ga) <= dm_rung_Y(gb)} and at least one of the
    inequalities is strict. The result is {POS} (+1) if the same
    condition holds with {ga} and {gb} interchanged. The result is
    {ZER} (0) in all other cases, namely {ga == gb} or the X- and
    Y-orders are inconsistent. */
  
dm_rung_t dm_rung_throw(dm_rung_t *glo, dm_rung_t *ghi, bool_t diagonal);
  /* Selects a rung at random whose X- and Y-coordinates lie between
    those of {glo} and {ghi}. If {diagonal}, the point will be on the
    approximate diagonal of that rectangular region.
    (TO FIX!!! {diagonal} should be a probability of insert/delete.) */

dm_rung_t dm_rung_throw_middle(dm_rung_t *glo, dm_rung_t *ghi, double skipProb);
  /* Selects a random rung {g} whose X- and Y-coordinates lie between
    those of {glo} and {ghi}, and such that {dm_rung_R(g)} is
    {(dm_rung_R(glo)+dm_rung_R(ghi))/2}, randomly rounded.
    
    The parameter {skipProb} specifies the ratio of axial steps to
    diagonal steps on a virtual path from {glo} to {ghi}; the rung {g}
    is taken to be the middle rung of that path. (The actual ratio may
    be greater than that, depending on {glo,ghi}.) */

vec_typedef(dm_rung_vec_t,dm_rung_vec,dm_rung_t);

dm_rung_vec_t dm_rung_vec_throw(dm_rung_t *gini, dm_rung_t *gfin, double skipProb);
  /* Generates a random unitary monotonic sequence of rungs {gv} 
    that starts with rung {gini} and ends with rung {gfin}.
    
    The parameter {skipProb} specifies the desired ratio of
    axial steps to diagonal steps. (The actual ratio may 
    be greater than that, depending on {gini,gfin}.) */

/* STEPS

  A /step/ is a pair of rungs (intended to be between the same two sequences).
  
  A step {g -> h} is /unitary/ if each coordinate changes by at most {1}.
  
  A step {g -> h} is /monotonic/ if no coordinate decreases, and at
  least one of them increases: that is, {g.X <= h.X} and {g.Y <= h.Y}
  but {g != h}.
  
  A step {g -> h} is `diagonal' (a `hop') if the X and Y coordinates
  change by the same amount; that is, {h.X - g.X == h.Y - g.Y}. 
  The step is `axial' (a `skip') if exactly one of the coordinates
  remains constant; that is, {g.X == h.X} or {g.Y == h.Y}. */
  
bool_t dm_rung_step_is_monotonic(dm_rung_t *g, dm_rung_t *h, bool_t die);
  /* TRUE iff the step from rung {g} to rung {h} is monotonic. 
    If {die} is true, and the pairing is not monotonic, fails
    with an error message. If {die} is FALSE, returns TRUE or FALSE
    normally without any messages. */

bool_t dm_rung_step_is_unitary_monotonic(dm_rung_t *g, dm_rung_t *h, bool_t die);
  /* TRUE iff the step from rung {g} to rung {h} is unitary
    and monotonic. If {die} is true, and the pairing is not valid, fails
    with an error message. If {die} is FALSE, returns TRUE or FALSE
    normally without any messages. */

void dm_rung_bresenham(dm_rung_t g, dm_rung_t h, int *ngP, dm_rung_vec_t *gv);
  /* Appends to {gv[0..*ngP-1]} a unitary monotonic series of rungs
    that interpolate between rung {gini} (exclusive) and rung {gfin}
    (inclusive), in the straightest possible way -- that is, a
    Bresenham path on the X-Y grid. Increments {*ngP} and expands {*gv}
    as needed. */

typedef dm_rung_t dm_rung_map_proc_t(dm_rung_t g);
  /* A client procedure that applies an arbitrary transformation
    to a rung {g}. */

dm_rung_vec_t dm_rung_vec_map_coordinates(dm_rung_vec_t *gv, dm_rung_map_proc_t *map);
  /* Given a rung list {gv}, returns a new rung list {q}
    such that {q[i] = map(gv[i])} for all valid {i}. */

dm_rung_vec_t dm_rung_vec_map_to_finer
  ( dm_rung_vec_t *gv, 
    int nxold, 
    int nyold, 
    int nxnew, 
    int nynew, 
    int nwtb
  );
  /* Applies {dm_seq_index_map_to_finer(ix,nxold,nxnew,nwtb)} to every
    X-coordinate {ix} in {gv}, and {dm-seq_index_map_to_finer(iy,nyold,nynew,nwtb)} to every
    Y-coordinate {iy}. If {gv} is monotonic, the result is monotonic. */

dm_rung_vec_t dm_rung_vec_map_to_coarser
  ( dm_rung_vec_t *gv, 
    int nxold, 
    int nyold, 
    int nxnew, 
    int nynew, 
    int nwtb
  );
  /* Applies {dm_seq_index_map_to_coarser(ix,nxold,nxnew,nwtb)} to
    every X-coordinate {ix} in {gv}, and
    {dm-seq_index_map_to_coarser(iy,nyold,nynew,nwtb)} to every
    Y-coordinate {iy}. If {gv} is monotonic, the result should be
    monotonic, except possibly for null steps. */

/* RUNG/STEP I/O */

void dm_rung_step_write(FILE *wr, dm_rung_t *g, dm_rung_t *h);
  /* Writes the step {g->h} in the format "{DX},{DY}|" where {DX} and
    {DY} are the increments in each coordinate. The following special
    cases apply:
    
      "|"    means "1,1|"     (a diagonal unit step)
      "/"    means "1,0|"     (a +1 step in X only)
      "\"    means "0,1|"     (a +1 step in Y only)
      "{D}|" means "{D},{D}|" 
      "{D}/" means "{D},0|" 
      "{D}\" means "0,{D}|" */

dm_rung_t dm_rung_step_read(FILE *rd, dm_rung_t g);
  /* Reads from the file {rd} a single step of a pairing,
    as encoded by {dm_pairing_step_write}. The step increments
    {DX,DY} are applied to the given starting rung {g} to produce 
    the final rung {h}. */

#endif

