#ifndef dm_score_H
#define dm_score_H

/* Scoring functions for rungs and steps of pairings. */
/* Last edited on 2006-12-25 23:16:12 by stolfi */

#define dm_score_H_COPYRIGHT \
  "Copyright  2006  by the State University of Campinas (UNICAMP)" \
  " and the Federal Fluminense University (UFF)"

#include <dm_basic.h>
#include <dm_datum.h>
#include <dm_seq.h>
#include <dm_rung.h>

#include <argparser.h>

typedef struct dm_score_args_t
  { double eql;      /* Score increment (us. positive) for a rung between equal raw datums. */
    double dif;      /* Score increment (us. negative) for a rung between distinct raw datums. */
    bool_t discrete; /* TRUE: {eql,dif} apply at rungs, FALSE apply along step. */
    double skp;      /* Score increment (us. negative) for an asymmetric step. */
  } dm_score_args_t; 
  /* A scoring criterion. */

/* RUNG AND STEP SCORES */

double dm_score_datum_rung(double eql, double dif, dm_datum_t *dx, dm_datum_t *dy);
  /* Computes the score of a rung {g} that pairs a datum {dx} with a datum{dy},
    according to the scoring weights {eql} and {dif}.  */

double dm_score_datum_step
  ( double eql, 
    double dif,
    dm_datum_t *d0x, 
    dm_datum_t *d0y, 
    dm_datum_t *d1x,
    dm_datum_t *d1y
  );
  /* Computes the mean rung score (as defined by
    {dm_score_datum_rung}) of a rung {t} that pairs a datum
    {dx}, that interpolates linearly between {d0x} and {d1x}, with a
    datum {dy}, that interpolates linearly between {d0y} and {d1y}.
    Uses the scoring weights {eql} and {dif}. */

double dm_score_step_increments(double skp, int kx, int ky);
  /* Computes the assymetry score of a step {g -> h} that advances
    {kx} units in the X-coordinate and {ky} units in the Y-coordinate,
    according to criterion {sc}.  The step should be unitary monotonic. */

double dm_score_rung(dm_score_args_t *sc, dm_rung_t g, dm_seq_t *xp, dm_seq_t *yp);
  /* Computes the score of a rung {g} between sequences {*xp} and
    {*yp}, according to the scoring criterion {sc}. 
    
    If {sc.discrete} is TRUE, returns
    {dm_score_datum_rung(sc.eql,sc.dif,dx,dy)} where {dx,dy} are the
    datums of {xp} and {yp}, respectively, that are associated by {g}.
    If {sc.discrete} is FALSE, returns zero. */

double dm_score_step(dm_score_args_t *sc, dm_rung_t g, dm_rung_t h, dm_seq_t *xp, dm_seq_t *yp);
  /* Computes the score of a step {g -> h} between sequences {*xp} and
    {*yp}, according to the scoring criterion {sc}.  The step 
    should be unitary monotonic.
    
    The result is the sum of a /datum term/ {Sd} that reflects the similarity
    between the datums paired along the step, and a /skip term/
    {Sk} which measures the asymmetry of the step {g -> h}.
    
    If {sc.discrete} is TRUE, the datum term {Sd} is just the
    {dm_score_datum_rung(sc.eql,sc.dif,dx,dy)} applied to the datum
    pair {dhx,dhy} connected by {h}. The rung {g} is irrelevant. If
    {h} is {dm_rung_none}, the corresponding value of {F(dx,dy)} is
    taken to be zero.
    
    If {sc.discrete} is FALSE, {Sd} is the mean value of {F(dx,dy)}
    where {dx,dy} interpolate linearly between {dgx,dgy} and
    {dhx,dhy}. If either rung is {dm_rung_none}, the datum term {Sd}
    is zero.
    
    The skip term {Sk} is defined by {dm_score_step_increments(sc.skp,kx,ky)}
    where {kx,ky} are the increments along each axis from {g} to {h}.
    If either rung is {dm_rung_none}, the skip term {Sk} is zero. */

dm_score_args_t dm_score_args_parse(argparser_t *pp);
  /* Parses a scoring criterion {sc} from the command line argument
    parser {pp}, in the format described by {dm_score_args_HELP} and
    {dm_score_args_INFO} below. */

#define dm_score_args_HELP \
  "{ discrete | smooth } {EQSCORE} {DIFFSCORE} {SKIPSCORE}"

#define dm_score_args_INFO \
  "The parameters {EQSCORE} and {DIFSCORE} specify" \
  " the `data' part {SD} of the score, that depends on the" \
  " similarity of the elements being paired.  At" \
  " the finest (unfiltered) level, each rung between two" \
  " identical elements is worth {EQSCORE} points (which" \
  " must be positive), and each rung between unequal elements" \
  " is worth {DIFFSCORE} points (which must be" \
  " negative).  At coarser (filtered) scales, the" \
  " points are reduced in proportion to the" \
  " square of the sample values.\n" \
  "\n" \
  "    The \"discrete\" keyword, if present, specifies" \
  " that the {SD} term above should be" \
  " computed only at the rungs of the pairing.  If \"smooth\" is" \
  " present, the {SD} term is computed by" \
  " interpolating the paired elements linearly between" \
  " consecutive rungs, and integrating the corresponding points.\n" \
  "\n" \
  "    The parameter {SKIPSCORE} (also" \
  " negative)is the penalty to be added for each" \
  " skew step (insertion/deletion).  The sum" \
  " of these penalties over all steps of" \
  " the pairing is `skip' part {SK} of the score."

#endif
