#ifndef dm_seq_H
#define dm_seq_H

/* Filtered DNA sequences */
/* Last edited on 2007-01-04 01:42:10 by stolfi */

#define dm_seq_H_COPYRIGHT \
  "Copyright  2005  by the State University of Campinas (UNICAMP)" \
  " and the Federal Fluminense University (UFF)"

#include <dm_basic.h>
#include <dm_sample.h>
#include <dm_datum.h>
#include <dm_psplot.h>

#include <vec.h>
#include <pswr.h>

/* This interface defines a discrete representation for 
  signals, as strings of /datum/ elements ({dm_datum_t}s).
  
  In a sequence of order zero, there is one datum for each codon, and
  the densities are computed for that single codon (either +1 or -1).
  A sequence of order {r > 0} is obtained by smoothing the sequence of
  order 0 with a Gaussian filter of appropriate width, and sampling
  the result with step 2 starting at 0.5. */

typedef int dm_seq_id_t; 
  /* A {dm_seq_id_t} is an internal identifier for a sequence, 
     presently an index into a table of sequences. */

#define dm_seq_id_none -1
  /* A {dm_seq_id_t} value that means "no sequence". */

typedef struct dm_seq_t
  { dm_seq_id_t id;         /* Internal sequence id (index into table). */
    char *name;             /* External sequence id (key in databank). */
    char *cmt;              /* Comments (or NULL). */
    bool_t circ;            /* TRUE if the sequence is circular. */
    dm_datum_scale_t sfac;  /* Per-channel scale factors for decoded samples. */
    dm_datum_vec_t dv;      /* The datum array. */
  } dm_seq_t;  
  /* A multichannel signal, finite or circular, sampled and encoded. */
  
int dm_seq_length(dm_seq_t *seqp);
  /* Number of samples in sequence {seqp}, whether open or circular. */

bool_t dm_seq_is_circular(dm_seq_t *seqp);
  /* TRUE iff {seqp} is circular. */

dm_sample_t *dm_seq_get_sample_address(dm_seq_t *seqp, int i, int k);
dm_sample_t dm_seq_get_sample(dm_seq_t *seqp, int i, int k);
void dm_seq_set_sample(dm_seq_t *seqp, int i, int k, dm_sample_t s);
  /* Extracts (resp. sets) the sample in channel {k} of datum {i} from
    the sequence {*seqp}. If the sequence is circular, the index {i} is
    reduced modulo {seqp->n}; otherwise {i} must lie in the range
    {[0..seqp->dv.nel-1]}. */
 
dm_datum_t *dm_seq_get_datum_address(dm_seq_t *seqp, int i);
dm_datum_t dm_seq_get_datum(dm_seq_t *seqp, int i);
void dm_seq_set_datum(dm_seq_t *seqp, int i, dm_datum_t d);
  /* Extracts (resp. sets) datum {i} from the sequence {*seqp}. If the
    sequence is circular, the index {i} is reduced modulo {seqp->n};
    otherwise {i} must lie in the range {[0..seqp->n-1]}. */

dm_seq_t dm_seq_new(int n);
  /* Allocates a new datum sequence of scale {scale} with initial size 
    at least {n}.  The sequence will have {id=-1}, {name=cmt=NULL}, and 
    {circular=FALSE}. */

dm_seq_t dm_seq_from_datum_vec
  ( dm_seq_id_t id, 
    char *name, 
    char *cmt, 
    bool_t circular,
    dm_datum_scale_t *sfac,
    dm_datum_vec_t dv
  );
  /* Assembles a sequence descriptor ({dm_seq_t}) with the given fields. */
  
int dm_seq_filtered_size(int nold, bool_t circ, int nwtb);
  /* Size of the sequence that will be returned by {dm-seq_filter},
    given that the input sequence has {nold} elements, circularity {circ},
    and the filter table has {nwtb} entries.
    
    If {circ} is true, the result is {nold/2} elements, 
    otherwie it is {(nold - wtbp->nel + 1)/2} elements */

dm_seq_t dm_seq_filter(dm_seq_t *oldp, double_vec_t *wtbp, char *wcmt);
  /* Filters and resamples a numeric sequence at double step size.
    More precisely, datum {i} of the new sequence is  
    
      {SUM{ oldp[2*i+j] * wtbp[j] : j in 0..wtbp->nel-1}}
      
    Let {nold = dm_seq_length(oldp)}. If {oldp->circular} is true, the
    index {2*i+j} is taken modulo {nold}, otherwise the index {2*i+j}
    lies in {0..nold-1}.
    
    The new sequence will have the same attributes {id,name,circular}
    as {oldp}. For each channel [k], the new {sfac.f[k]} will be the
    root-mean-square value of the samples in channel {k]. The string
    {wcmt}, if not NULL, is assumed to be the name or description of
    the weight table {wtb}, and is appended to the comment {oldp->cmt}
    to make the new string's comment. */

int dm_seq_index_map_affinely(int ixold, int nold, int nnew);
int dm_seq_index_map_to_finer(int ixold, int nold, int nnew, int nwtb);
int dm_seq_index_map_to_coarser(int ixold, int nold, int nnew, int nwtb);
  /* Maps the index {ixold} of a datum in some sequence {sold}
    to the index {ixnew} of the approximately corresponding datum in 
    another sequence {snew}.  
    
    The parameters {nold} and {nnew} should be the datum counts of the
    two sequences.
    
    The procedure {dm_seq_index_map_affinely} does an affine map from
    range {0..nxold} to range {0..nxnew}, rounding the result to
    integer. The procedures {dm_seq_index_map_to_finer} and
    {dm_seq_index_map_to_coarser} assume that the coarser-scale
    sequence was obtained from the finer-scale sequence by filtering
    with a filter weight table of size {nwtb} and downsampling the
    result with a stride of 2.
    
    All three procedures are monotonic non-decreasing: i.e.
    increasing {ixold} does not decrease the result. */

/* RANDOM SEQUENCES */

void dm_seq_throw_segment(int ns, bool_t circ, int len, int *ini, int *fin);
  /* Selects initial and final sample indices {*ini,*fin} for a random 
    segment with {len} samples, taken from a sequence with {ns} samples
    and circularity {circ}. */ 

/* SEQUENCE I/O

  The procedures in this section read or write a datum sequence from
  a file. The file consists of 
  
    a standard header line, "begin numeric_bio_seq (version of ...)";
    
    zero or more comment lines, starting with "| ";
    
    a line "circular = {BOOL}";
    
    a line "scale = {SFACTOR} ...", where each {SFACTOR} is a fractional scale 
      factor to be multiplied into all samples of the corresponding channel,
      after applying {dm_sample_decode};      
    
    a line "elements = {NSMP}";
    
    {NSMP} encoded numeric datums, each {dm_CHANNEL} integers, one per line;
    
    a standard footer line, "end numeric_bio_seq".
  */

dm_seq_t dm_seq_read(FILE *rd);
  /* Reads a numeric DNA/RNA sequence from file {rd}.
    The sequence will have {id=dm_seq_id_none}, {name=NULL}. */
  
void dm_seq_write(FILE *wr, dm_seq_t *seqp);
  /* Writes a numeric DNA/RNA sequence to file {wr}. */

void dm_seq_write_named(dm_seq_t *seqp, char *name, char *tag);
  /* Same as {dm_seq_write}, but creates a disk file called"{name}{tag}.egs". */

/* DNA/RNA SEQUENCE I/O */

dm_seq_t dm_seq_from_nucleic_string(dm_seq_id_t id, char *name, char *cmt, char *bas, bool_t circ);
  /* Converts the DNA/RNA string {bas} into a {dm_datum_t} sequence.
    The sequence's numeric and alphabetic identifiers will be {id} and
    {name}, respectively. The comment text will be {txt}. The sequence
    is assumed to be circular iff {circ} is true. */

dm_seq_t dm_seq_read_from_nucleic_file(dm_seq_id_t id,  char *name, char *fname, bool_t circ);
  /* Reads a DNA/RNA sequence from file "{fname}.bas". The sequence's numeric and 
    alphabetic identifiers will be {id} and {name}, respectively.
    The sequence is assumed to be circular iff {circ} is true. */

/* POSTSCRIPT PLOTTING */

void dm_seq_postscript_plot(dm_psplot_t *dp, dm_seq_t *seqp);
  /* Generates a Postscript plot of the sequence {seqp}, written out
    to the Postscript plotting stream {dp}. The plot uses the whole
    plotting area of {dp} (with a bit of slack). Each channel is decoded and
    drawn with a different color. */

void dm_seq_postscript_plot_named
  ( dm_seq_t *seqp, 
    double hSize, 
    double vSize,
    double mrg,
    char *name,
    char *tag
  );
  /* Same as {dm_seq_postscript_plot} but writes an EPS file called
    "{name}{tag}.eps". The figure will have {hSize} by {vSize}
    millimeters, including a blank margin {mrg} millimeters wide all
    around. */

/* AUXILIARY FUNCTIONS */

void dm_double_vec_filter(double_vec_t *oldp, double_vec_t *wtbp, double_vec_t *newp);
  /* Computes the convolution of the sample sequence {oldp} with the
    weight sequence {wtbp}, displaced by two steps at a time, and
    stores the result in {newp}. More precisely, sets {newp[i]} to 
    
      {SUM{ oldp[2*i+j] * wtbp[j] : j in 0..wtbp->nel-1}}
      
    where the index {2*i+j} is taken modulo {oldp->nel}. */

#endif
