#define PROG_NAME "dm_test_distance"
#define PROG_DESC "test of sample distance functions"
#define PROG_VERS "1.0"

/* Last edited on 2007-01-04 00:11:57 by stolfi */

#define dm_test_distance_C_COPYRIGHT \
  "Copyright  2006  by the State University of Campinas (UNICAMP)"

#define PROG_HELP \
  PROG_NAME " \\\n" \
  "  -trials {N_TRIALS} \\\n" \
  "  -filterWeights " dm_weight_args_HELP " \\\n" \
  "  {OUT_NAME}"

#define PROG_INFO \
  "NAME\n" \
  "  " PROG_NAME " - " PROG_DESC "\n" \
  "\n" \
  "SYNOPSIS\n" \
  "  " PROG_HELP "\n" \
  "\n" \
  "DESCRIPTION\n" \
  "  This program tests distance estimators" \
  " under various filter and distance weights.\n" \
  "OUTPUT FILES\n" \
  "  All output files will have names starting with {OUT_NAME}.\n" \
  "\n" \
  "OPTIONS\n" \
  "  -trials {N_TRIALS}\n" \
  "    This mandatory argument specifies the" \
  " number of test cases to generate.\n" \
  "\n" \
  "  -filterWeights " dm_weight_args_HELP "\n" \
  "    This mandatory argument specifies the" \
  " weights to be used to compute the smoothed" \
  " signals.  " dm_weight_args_norm_sum_INFO ".\n" \
  "\n" \
  argparser_help_info_HELP_INFO "\n" \
  "SEE ALSO\n" \
  "  dm_match(1)\n" \
  "\n" \
  "AUTHOR\n" \
  "  This program was created on 21/dec/2006 by J. Stolfi.\n" \
  "WARRANTY\n" \
  argparser_help_info_NO_WARRANTY "\n" \
  "\n" \
  "RIGHTS\n" \
  "  " dm_test_distance_C_COPYRIGHT ".\n" \
  "\n" \
  argparser_help_info_STANDARD_RIGHTS  

#include <dm_basic.h>
#include <dm_seq.h>
#include <dm_nucleic.h>
#include <dm_weight.h>

#include <jsrandom.h>
#include <bool.h>
#include <affirm.h>
#include <argparser.h>
#include <gauss_elim.h>

#include <math.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

typedef struct dm_options_t 
  { double_vec_t w;    /* Filter weights. */
    int trials;        /* Number of test cases to generate. */
    /* Output parameters: */
    char *outName;     /* Output file name prefix (minus extensions). */
  } dm_options_t;
  
int main(int argc, char**argv);

#define dm_BASIS_FUNCTIONS 3

void dm_td_fit_dist_formulas(double_vec_t *w, int nt, char *outName);
  /* Estimates the optimum distance formula for the weight vector {w},
    using {nt} test cases. */

void dm_td_generate_case(double *w, int nw, double *X, double *Y, double *Z, bool_t verbose);
/* Generates two random nucleotide strings with {nw} letters, where
  {nw = w->nel}.  Computes their numeric equivalents {x[0..nw-1]}
  and {y[0..nw-1]}, where each {x[i]} or {y[i]} is a 3-vector 
  with the convention described under {dm_datum_decoded_from_nucleic_char}.
  Then computes the filter outputs {X = SUM{ w[i]*x[i] : i in 0..nw-1}}
  and similarly for Y.  Also computes {Z[i] = |x[i]-y[i]|^2} for 
  {i} in {0..nw-1}, and {Z[nw] = SUM{w[i]*Z[i] : i in 0..nw-1}}. */

void dm_td_apply_filter(int nw, double *w, int *x, double *X);
/* Applies a filter with weights {w[0..nw-1]} to the columns of 
  the integer array {x}.  Assumes that {x} has {nw} rows and {nc} columns,
  where {nc=dm_CHANNELS}. Returns the result in the float vector 
  {X[0..nc-1]}. */

void dm_normalize_doubles(double *v, int nv);
  /* Scales {v[0..nv-1]} so that they add to 1. */

double dm_doubles_normsq(double *v, int nv);
  /* Returns the sum of the squares of the values {v[0..nv-1]}. */

double dm_lsq_basis(int ib, double *X, double *Y);
  /* Evaluates basis function number {ib} on {X} and {Y},
    for {ib} in {0..dm_BASIS_FUNCTIONS-1}. */

char *dm_lsq_basis_txt(int ib);
  /* Retursn a description of basis function number {ib},
    for {ib} in {0..dm_BASIS_FUNCTIONS-1}. */

void dm_debug_double_vec(double *v, int nv, char *fmt);
void dm_debug_int_vec(int *v, int nv, char *fmt);
  /* These procedures print {v[0..nv-1]} to {stderr}, each with format {fmt},
    separated by spaces and bracketed by '[' and ']'. */

dm_options_t *dm_get_options(int argc, char**argv);
  /* Parses the command line options, packs 
    them into a {dm_options_t} record. */

int main(int argc, char**argv)
  { 
    dm_options_t *o = dm_get_options(argc, argv);
    
    dm_td_fit_dist_formulas(&(o->w), o->trials, o->outName);

    return 0;
  }
  
void dm_td_fit_dist_formulas(double_vec_t *wtb, int nt, char *outName)
  { 
    fprintf(stderr, "----------------------------------------------------------------------\n");
    fprintf(stderr, "Testing distances...\n");
    fprintf(stderr, "...\n");

    /* Get problem dimensions: */
    int nw = wtb->nel;          /* Number of entries in weight table. */
    int nc = dm_CHANNELS;       /* Number of channels in datum. */
    int nz = nw+1;              /* Number of functions to approximate. */
    int nb = (nw == 1 ? 2 : 3); /* Number of relevant basis functions. */
    
    /* Print parameters and weight vector: */
    fprintf(stderr, "nt = %d\n", nt);
    fprintf(stderr, "nw = %d\n", nw);

    /* Normalize weights to unit sum: */
    double w[nw];
    int iw; for (iw = 0; iw < nw; iw++) { w[iw] = wtb->el[iw]; }
    dm_normalize_doubles(w, nw);
    fprintf(stderr, "w = ");
    dm_debug_double_vec(w, nw, "%10.8f"); 
    fprintf(stderr, " var = %12.10f", dm_doubles_normsq(w, nw));
    fprintf(stderr, "\n");
    
    /* Allocate storage for independent variables: */
    double X[nc];  /* Filter output for X sequence. */
    double Y[nc];  /* Filter output for Y sequence. */ 

    /* Allocate storage for dependent variables: */
    double Z[nz];  /* {Z[i] = |x[i]-y[i]|^2} if {i<nw}, else {D2(x,y)}. */
    
    /* LSq fitting system {A U = B} for {Z[0..nz-1]} in terms of {phi[0..nb-1](X,Y)}: */
    double A[nb*nb];  /* Has {nb} rows and {nb} columns. */
    double B[nb*nz];  /* Has {nb} rows and {nz} columns. */
    int k;
    for (k = 0; k < nb*nb; k++) { A[k] = 0.0; }
    for (k = 0; k < nb*nz; k++) { B[k] = 0.0; }
    
    /* Generate all test cases, accumulate statistics: */
    int it;
    for (it = 0; it < nt; it++)
      { 
        bool_t verbgen = (it < 2); /* Debug the test case generator? */
        bool_t verbacc = (it < 20); /* Debug the stats accumulator? */
        
        /* Compute {d2[0..nw-1},X,Y} for a random seq pair: */
        dm_td_generate_case(w, nw, X, Y, Z, verbgen);
        if (verbacc) 
          { fprintf(stderr, "  X =");
            dm_debug_double_vec(X, nc, "%6.3f");
            fprintf(stderr, "  Y =");
            dm_debug_double_vec(Y, nc, "%6.3f");
            fprintf(stderr, "  Z =");
            dm_debug_double_vec(Z, nz, "%6.3f");
            fprintf(stderr, "\n");
          }
        
        /* Compute the required basis functions of {X,Y}: */
        double phi[nb]; 
        int ib;
        for (ib = 0; ib < nb; ib++) { phi[ib] = dm_lsq_basis(ib,X,Y); }
        if (verbacc)
          { fprintf(stderr, "  phi = ");
            dm_debug_double_vec(phi, nb, "%9.5f");
            fprintf(stderr, "\n");
          }
        
        /* Accumulate scalar products on matrix: */
        for (ib = 0; ib < nb; ib++)
          { int jb, jz;
            for (jb = 0; jb < nb; jb++)
              { A[ib*nb+jb] += phi[ib]*phi[jb]; }
            for (jz = 0; jz < nz; jz++)
              { B[ib*nz+jz] += phi[ib]*Z[jz]; }
          }
          
        if (verbacc) { fprintf(stderr, "\n"); }
      }
      
    /* Print and solve the least squares system: */
    fprintf(stderr, "Least squares systems:\n");
    int ib;
    for (ib = 0; ib < nb; ib++)
      { fprintf(stderr, "  %-4s", (ib == nb/2 ? "A = " : ""));
        dm_debug_double_vec(&(A[ib*nb]), nb, "%12.5f");
        fprintf(stderr, "  %-4s", (ib == nb/2 ? "B = " : ""));
        dm_debug_double_vec(&(B[ib*nz]), nz, "%12.5f");
        fprintf(stderr, "\n");
      }
          
    double U[nb*nw];
    int rank = gsel_solve(nb, nb, A, nz, B, U);
    demand(rank == nb, "could not solve the system");
    
    /* Print the fitted formula: */
    int iz;
    for (iz = 0; iz < nz; iz++)
      { fprintf(stderr, "formula for Z[%d]:\n", iz);
        int ib;
        for (ib = 0; ib < nb; ib++)
          { char *op = (ib == nb-1 ? "" : " +");
            fprintf(stderr, "  %+12.8f*(%s)%s\n", U[ib*nz + iz], dm_lsq_basis_txt(ib), op);
          }
        fprintf(stderr, "\n");
      }
    fprintf(stderr, "----------------------------------------------------------------------\n");
  }
    
void dm_debug_double_vec(double *v, int nv, char *fmt)
  { fprintf(stderr, "[");
    int j;
    for (j = 0; j < nv; j++)
      { fprintf(stderr, " "); fprintf(stderr, fmt, v[j]); }
    fprintf(stderr, " ]");
  }
    
void dm_debug_int_vec(int *v, int nv, char *fmt)
  { fprintf(stderr, "[");
    int j;
    for (j = 0; j < nv; j++)
      { fprintf(stderr, " "); fprintf(stderr, fmt, v[j]); }
    fprintf(stderr, " ]");
  }

double dm_lsq_basis(int ib, double *X, double *Y)
  { switch(ib)
      { case 0: 
          return 1;
        case 1: 
          return X[0]*Y[0] + X[1]*Y[1] + X[2]*Y[2];
        case 2: 
          return X[0]*X[0] + X[1]*X[1] + X[2]*X[2] + Y[0]*Y[0] + Y[1]*Y[1] + Y[2]*Y[2];
        default: 
          affirm(FALSE, "bad ib"); return 0;
      }
  }
    
char *dm_lsq_basis_txt(int ib)
  { switch(ib)
      { case 0: return "1";
        case 1: return "<X|Y>";
        case 2: return "(|X|^2 + |Y|^2)";
        default: affirm(FALSE, "bad ib"); return "BAD";
      }
  }

void dm_td_generate_case(double *w, int nw, double *X, double *Y, double *Z, bool_t verbose)
  {      
    /* Get problem dimensions: */
    int nc = dm_CHANNELS; /* Number of channels in datum. */
    
    /* DNA alphabet: */
    char *atcg = "ATCG"; /* The four nucleotides. */
    
    /* The two DNA letter sequences: */
    char xc[nw + 1]; xc[nw] = 0; /* First nucleotide string (with final 0). */
    char yc[nw + 1]; yc[nw] = 0; /* Second nucleotide string (with final 0). */

    /* The numeric coordinates of {xc[0..nw-1],yc[0..nw-1]}: */
    int x[nw*nc];  /* Element {x[i*nc+c]} is channel {c} of numerized {xc[i]}. */ 
    int y[nw*nc];  /* Element {y[i*nc+c]} is channel {c} of numerized {yc[i]}. */

    /* Generate two random nucleotide strings {x,y}: */
    randchars(xc, nw, atcg, 4);
    randchars(yc, nw, atcg, 4);
    if (verbose) 
      { fprintf(stderr, "  x = %s", xc);
        fprintf(stderr, "  y = %s", yc);
        fprintf(stderr, "\n"); 
      }
    /* Convert them to numeric vectors: */
    int iw;
    for (iw = 0; iw < nw; iw++) 
      { dm_datum_decoded_from_nucleic_char(xc[iw], &(x[iw*nc]));
        dm_datum_decoded_from_nucleic_char(yc[iw], &(y[iw*nc]));
        if (verbose)
          { fprintf(stderr, "  x[%02d] = %c = ", iw, xc[iw]); 
            dm_debug_int_vec(&(x[iw*nc]), nc, "%+d");
            fprintf(stderr, "  y[%02d] = %c = ", iw, yc[iw]); 
            dm_debug_int_vec(&(y[iw*nc]), nc, "%+d");
            fprintf(stderr, "\n"); 
          }
      }
    /* Compute the decoded non-quantized filtered datums {X,Y}: */
    int c;
    for (c = 0; c < nc; c++)
      { dm_td_apply_filter(nw, w, x, X);
        dm_td_apply_filter(nw, w, y, Y);
      }
    if (verbose)
      { fprintf(stderr, "  X = "); 
        dm_debug_double_vec(X, nc, "%+10.6f");
        fprintf(stderr, "  Y = "); 
        dm_debug_double_vec(X, nc, "%+10.6f");
        fprintf(stderr, "\n"); 
      }
    
    /* Compute target quantities {Z[i]}: */
    double sumd2 = 0;
    for (iw = 0; iw < nw; iw++)
      { double sumd2i = 0;
        for (c = 0; c < nc; c++)
          { int d = x[iw*nc+c] - y[iw*nc+c]; sumd2i += d*d; }
        Z[iw] = sumd2i;
        sumd2 += sumd2i*w[iw];
      }
    Z[nw] = sumd2;
  }

void dm_normalize_doubles(double *v, int nv)
  { double sum = 0;
    int i;
    for (i = 0; i < nv; i++) { sum += v[i]; }
    if (sum == 0) { return; }
    for (i = 0; i < nv; i++) { v[i] /= sum; }
  }

double dm_doubles_normsq(double *v, int nv)
  { int i;
    double sum2 = 0;
    for (i = 0; i < nv; i++) { double d = v[i]; sum2 += d*d; }
    return sum2;
  }

void dm_td_apply_filter(int nw, double *w, int *x, double *X)
  { int nc = dm_CHANNELS;
    int c;
    for (c = 0; c < nc; c++)
      { double sumwx = 0;
        int iw;
        for (iw = 0; iw < nw; iw++)
          { sumwx += w[iw]*((double)x[iw*nc+c]); }
        X[c] = sumwx;
      }
  }

dm_options_t *dm_get_options(int argc, char**argv)
  { 
    dm_options_t *o = (dm_options_t *)notnull(malloc(sizeof(dm_options_t)), "no mem");
    
    argparser_t *pp = argparser_new(stderr, argc, argv);
    argparser_set_help(pp, PROG_HELP);
    argparser_set_info(pp, PROG_INFO);
    argparser_process_help_info_options(pp);
    
    argparser_get_keyword(pp, "-trials");
    int ntmin = dm_BASIS_FUNCTIONS; /* Lsq needs at least this many data points */
    o->trials = argparser_get_next_double(pp, ntmin, INT_MAX);
    
    argparser_get_keyword(pp, "-filterWeights");
    o->w = dm_weight_args_parse(pp, TRUE);
    
    argparser_skip_parsed(pp);
    
    o->outName = argparser_get_next(pp);

    argparser_finish(pp);
    
    return o;
  }
