#define PROG_NAME "dm_test_mapping"
#define PROG_DESC "test of mapping candidates between scales"
#define PROG_VERS "1.0"

/* Last edited on 2007-01-04 01:56:07 by stolfi */

#define dm_test_mapping_C_COPYRIGHT \
  "Copyright  2006  by the State University of Campinas (UNICAMP)"

#define PROG_HELP \
  PROG_NAME " \\\n" \
  "  {SEQ_LENGTH_X} {CIRC_X} \\\n" \
  "  {SEQ_LENGTH_Y} {CIRC_Y} \\\n" \
  "  {N_CANDS} \\\n" \
  "  {OUT_NAME}"

#define PROG_INFO \
  "NAME\n" \
  "  " PROG_NAME " - " PROG_DESC "\n" \
  "\n" \
  "SYNOPSIS\n" \
  "  " PROG_HELP "\n" \
  "\n" \
  "DESCRIPTION\n" \
  "  This program generates {N_CANDS} candidate pairings" \
  " between two hypothetical biosequences with lengths" \
  " {SEQ_LENGTH_X} and {SEQ_LENGTH_Y}, reduced by" \
  " a factor {2^M} for some {M}, and then maps" \
  " those candidates at succesively finer scales.\n" \
  "\n" \
  "OUTPUT FILES\n" \
  "  All output files will have names starting with {OUT_NAME}.\n" \
  "\n" \
  "OPTIONS\n" \
  "  None.\n" \
  "\n" \
  argparser_help_info_HELP_INFO "\n" \
  "SEE ALSO\n" \
  "  dm_match(1)\n" \
  "\n" \
  "AUTHOR\n" \
  "  This program was created on 17/dec/2006 by J. Stolfi.\n" \
  "WARRANTY\n" \
  argparser_help_info_NO_WARRANTY "\n" \
  "\n" \
  "RIGHTS\n" \
  "  " dm_test_mapping_C_COPYRIGHT ".\n" \
  "\n" \
  argparser_help_info_STANDARD_RIGHTS  

#include <dm_basic.h>
#include <dm_seq.h>
#include <dm_image.h>
#include <dm_pairing.h>
#include <dm_cand.h>

#include <affirm.h>
#include <float_image.h>
#include <float_pnm_image.h>
#include <jspnm_image.h>
#include <argparser.h>

#include <math.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

#define dm_MAX_LEVELS 20
  /* Maximum levels to consider above level 0. */
  
#define dm_MIN_SEQ_SIZE 20
  /* Minimum number of datums in a sequence at coarsest level. */

typedef struct dm_options_t 
  { int seqLengthX;    /* length of sequence {A}. */
    bool_t circX;      /* TRUE iff sequence {A} is circular. */
    int seqLengthY;    /* length of sequence {B}. */
    bool_t circY;      /* TRUE iff sequence {B} is circular. */
    /* Filtering parameters: */
    int nCands;        /* Number of candidates to generate. */
    /* Output parameters: */
    char *outName;     /* Output file name prefix (minus extensions). */
  } dm_options_t;
  
int main(int argc, char**argv);

dm_options_t *dm_get_options(int argc, char**argv);
  /* Parses the command line options, packs 
    them into a {dm_options_t} record. */

void dm_compute_multiscale_lengths(int ns, bool_t circ, int maxLevel, int nwt[], int nfs[]);
 /* Computes the lengths of sequeneces produces by {dm_MAX_LEVELS}
   successive applications of {dm_seq_filter} on a sequence with {ns}
   elements. 
   
   Assumes that the fitering from level {r} to level {r+1} uses a
   weight table with {nwt[r]} entries. If {circ} is TRUE, assumes that
   the original sequence and all its refined versions are circular.
   
   The computed lengths are returned in {nfs[0..dm_MAX_LEVELS]}, with
   {nfs[0] = ns}. */

dm_cand_vec_t dm_fake_initial_cands(dm_options_t *o, int nx, int ny);
  /* Generates a list of {o->nCands} candidates for the two {dm_seq_t}s.
    Assumes that the first sequence has {nx} elements, and is circular
    iff {o->circX} is TRUE. Ditto for {ny,o->circY}. */
  
int main(int argc, char**argv)
  { 
    int M = dm_MAX_LEVELS;
    
    dm_options_t *o = dm_get_options(argc, argv);

    fprintf(stderr, "choosing the filter window size for scales %d .. %d ...\n", 0, M-1);
    int nwt[M]; /* Number of weights in filter kernel at each scale. */
    int level;
    for (level = 0; level < M; level++) { nwt[level] = 7 + 2*(level % 3); }
    
    fprintf(stderr, "computing seq lengths at scales %d .. %d ...\n", 0, M);
    int nfx[M + 1]; dm_compute_multiscale_lengths(o->seqLengthX, o->circX, M, nwt, nfx);
    int nfy[M + 1]; dm_compute_multiscale_lengths(o->seqLengthY, o->circY, M, nwt, nfy);
    for (level = 0; level <= M; level++)
      { int nx = nfx[level], ny = nfy[level]; 
        fprintf(stderr, "  level %2d  seq lengths %3d %3d", level, nx, ny);
        if (level < M) 
          { int nw = nwt[level];
            fprintf(stderr, " filter window width %3d", nw);
          }
        fprintf(stderr, "\n");
      }

    fprintf(stderr, "reducing max level to ensure nonzero strings...\n");
    while ((M > 0) && ((nfx[M] < dm_MIN_SEQ_SIZE) || (nfy[M] < dm_MIN_SEQ_SIZE))) { M--; }
    fprintf(stderr, "  max level = %d\n", M);
    
    fprintf(stderr, "finding initial candidates at level %d ...\n", M);
    int naini = nfx[M]; fprintf(stderr, " seq 0 length = %d\n", naini);
    int nbini = nfy[M]; fprintf(stderr, " seq 1 length = %d\n", nbini);
    dm_cand_vec_t cdv = dm_fake_initial_cands(o, naini, nbini);

    fprintf(stderr, "multiscale mapping ...\n");
    for (level = M; level >= 0; level--)
      { 
        int nx = nfx[level];
        int ny = nfy[level];
        int nw = nwt[level];
        
        if (level < M)
          { /* Map candidates in {cdv} to the next finer scale: */
            fprintf(stderr, "mapping from level %d to level %d ...\n", level+1, level);
            fprintf(stderr, "  sequence 0 length %d -> %d\n", nfx[level+1], nx);
            fprintf(stderr, "  sequence 1 length %d -> %d\n", nfy[level+1], ny);
            fprintf(stderr, "  filter window size %d\n", nw);
            cdv = dm_cand_vec_map_to_finer(&cdv, nx, ny, nw);
          }
          
        /* Make filename prefix for current level: */
        char *levName = NULL;
        asprintf(&levName, "%s-%02d", o->outName, level);
        
        /* Write and plot candidates: */
        fprintf(stderr, "writing current candidates ...\n");
        dm_cand_vec_write_named(&cdv, levName, "-cd");
        dm_cand_vec_image_write_named(&cdv, nx, o->circX, ny, o->circY, levName, "-cd");
      }

    return 0;
  }
    
void dm_compute_multiscale_lengths(int ns, bool_t circ, int maxLevel, int nwt[], int nfs[])
  { int level;
    nfs[0] = ns;
    for (level = 1; level <= maxLevel; level++) 
      { int nold = nfs[level-1];
        int nwtb = nwt[level-1];
        int nnew = dm_seq_filtered_size(nold, circ, nwtb);
        if (nnew < 0) { nnew = 0; }
        nfs[level] = nnew;
      }
  }

dm_cand_vec_t dm_fake_initial_cands(dm_options_t *o, int nx, int ny)
  { /* Define the max candidate length {maxlen}: */
    int maxlen = (nx > ny ? nx : ny);
    int minlen = (maxlen + 9)/10;
    /* Generate the candidates: */
    dm_cand_vec_t cdv = dm_cand_vec_throw
      ( o->nCands,
        0, "a", nx, o->circX, 
        1, "b", ny, o->circY,
        minlen, maxlen,
        /*circProb*/ 0.1, 
        /*diagProb*/ 0.2,
        /*skipProb*/ 0.1
      );
    return cdv;
  }

dm_options_t *dm_get_options(int argc, char**argv)
  { 
    dm_options_t *o = (dm_options_t *)notnull(malloc(sizeof(dm_options_t)), "no mem");
    
    argparser_t *pp = argparser_new(stderr, argc, argv);
    argparser_set_help(pp, PROG_HELP);
    argparser_set_info(pp, PROG_INFO);
    argparser_process_help_info_options(pp);
    
    argparser_skip_parsed(pp);
    
    o->seqLengthX = argparser_get_next_int(pp, 0, INT_MAX);
    o->circX = argparser_get_next_bool(pp);
    
    o->seqLengthY = argparser_get_next_int(pp, 0, INT_MAX);
    o->circY = argparser_get_next_bool(pp);
    
    o->nCands = argparser_get_next_int(pp, 0, INT_MAX);
    o->outName = argparser_get_next(pp);

    argparser_finish(pp);
    
    return o;
  }
