#define PROG_NAME "dm_test_refining"
#define PROG_DESC "test of candidate refinement routines"
#define PROG_VERS "1.0"

/* Last edited on 2007-01-04 01:56:15 by stolfi */

#define dm_test_refining_C_COPYRIGHT \
  "Copyright  2006  by the State University of Campinas (UNICAMP)"

#define PROG_HELP \
  PROG_NAME " \\\n" \
  "  -nBases {N_BASES} \\\n" \
  "  [ -circular {CIRC_X} {CIRC_Y} ] \\\n" \
  "  -mutProb {MUT_PROB} \\\n" \
  "  -insProb {INS_PROB} \\\n" \
  "  -scores " dm_score_args_HELP " \\\n" \
  "  [ -delta {DELTA} ] [ -kappa {KAPPA} ] \\\n" \
  "  -nCands {N_CANDS} \\\n" \
  "  [ -repeat {N_TIMES} ] \\\n" \
  "  {OUT_NAME}"

#define PROG_INFO \
  "NAME\n" \
  "  " PROG_NAME " - " PROG_DESC "\n" \
  "\n" \
  "SYNOPSIS\n" \
  "  " PROG_HELP "\n" \
  "\n" \
  "DESCRIPTION\n" \
  "  This program generates {N_CANDS} random candidate pairings" \
  " between two random biosequences with approximately {N_BASES}" \
  " nucleotides, and then refines them, {N_TIMES} times.  The" \
  " sequences may be open or circular.\n" \
  "OUTPUT FILES\n" \
  "  All output files will have names starting with {OUT_NAME}.\n" \
  "\n" \
  "  All candidates are plotted together as image files" \
  " \"{OUT_NAME}-{XXX}-cd.pgm\" where {XXX} is \"ini\" (before" \
  " refinement) or \"fin\" (after it)." \
  "\n" \
  "  The base-to-base distance matrix for the two sequences" \
  " is plotted as file \"{OUT_NAME}-eq.pgm\"." \
  "\n" \
  "  The dynamic programming tableau for each candidate" \
  " and each iteration of the refinement procedure is plotted as image" \
  " file \"{OUTNAME}-{NNNNN}-{T}.pgm\" where {NNNNN} is a" \
  " five-digit candidate number, and {T} is a single-digit" \
  " iteration number.\n" \
  "\n" \
  "OPTIONS\n" \
  "  -nBases {N_BASES}\n" \
  "    This mandatory argument specifies the" \
  " approximate length (number of nucleotide bases)" \
  " of the two test sequences.\n" \
  "\n" \
  "  -circular {CIRC_X} {CIRC_Y}\n" \
  "    This optional argument specifies whether" \
  " the sequences are open or circular.  If {CIRC_X}" \
  " is \"T\" (true), the first (X) sequence will be" \
  " circular like a bacterial chromosome or" \
  " plasmid.  If {CIRC_X} is \"F\" (false)," \
  " the X sequence will be open.  Ditto for {CIRC_Y}" \
  " and the Y sequence.  The default is" \
  " \"-circular F F\" (both open).\n" \
  "\n" \
  "  -mutProb {MUT_PROB}\n" \
  "    This mandatory argument specifies the" \
  " nucleotide mutation probability used when" \
  " generating the two sequences.\n" \
  "\n" \
  "  -delProb {DEL_PROB}\n" \
  "    This mandatory argument specifies the" \
  " nucleotide deletion or insertion probability used" \
  " when generating the two sequences.\n" \
  "\n" \
  "  -scores " dm_score_args_HELP "\n" \
  "    This mandatory argument specifies the" \
  " scoring of steps in a pairing.  " dm_score_args_INFO "\n" \
  "\n" \
  "  -delta {DELTA}\n" \
  "    This optional argument specifies the" \
  " amount of adjustment allowed for the X and Y coordinates of" \
  " internal rungs of each pairing.  The default is 3.\n" \
  "\n" \
  "  -kappa {DELTA}\n" \
  "    This optional argument specifies the" \
  " amount of X and Y extension allowed" \
  " at either end of each pairing.  The default is 6.\n" \
  "\n" \
  "  -nCands {N_CANDS}\n" \
  "    This mandatory argument specifies how many candidates" \
  " to generate.\n" \
  "\n" \
  "  -repeat {N_TIMES}\n" \
  "    This optional argument specifies how many times each" \
  " candidate should be refined.  The default is 1.\n" \
  "\n" \
  argparser_help_info_HELP_INFO "\n" \
  "SEE ALSO\n" \
  "  dm_match(1)\n" \
  "\n" \
  "AUTHOR\n" \
  "  This program was created on 21/dec/2006 by J. Stolfi.\n" \
  "WARRANTY\n" \
  argparser_help_info_NO_WARRANTY "\n" \
  "\n" \
  "RIGHTS\n" \
  "  " dm_test_refining_C_COPYRIGHT ".\n" \
  "\n" \
  argparser_help_info_STANDARD_RIGHTS  

#include <dm_basic.h>
#include <dm_seq.h>
#include <dm_nucleic.h>
#include <dm_rung.h>
#include <dm_score.h>
#include <dm_pairing.h>
#include <dm_cand.h>
#include <dm_refine.h>
#include <dm_dyn.h>
#include <dm_image.h>
#include <dm_test_tools.h>

#include <affirm.h>
#include <float_image.h>
#include <float_pnm_image.h>
#include <jspnm_image.h>
#include <argparser.h>

#include <math.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

typedef struct dm_options_t 
  { /* Sequence generation parameters: */
    double nBases;       /* Expected number of bases. */
    bool_t circX;        /* Circularity of X sequence. */
    bool_t circY;        /* Circularity of Y sequence. */
    double mutProb;      /* Mutation probability. */
    double delProb;      /* Insertion/deletion probability. */
    /* Candidate generation parameters: */
    int nCands;          /* Number of candidates to generate. */
    /* Refinement iteration parameter: */
    int repeat;          /* How many times to refine each candidate. */
    /* Refinement parameters: */
    int delta;           /* Half-width of tableau around original pairing. */
    int kappa;           /* Extension of tableau beyond ends of pairing. */
    dm_score_args_t sc;  /* Scoring criterion. */
    /* Output parameters: */
    char *outName;       /* Output file name prefix (minus extensions). */
  } dm_options_t;
  
int main(int argc, char**argv);

dm_options_t *dm_get_options(int argc, char**argv);
  /* Parses the command line options, packs 
    them into a {dm_options_t} record. */

dm_cand_vec_t dm_fake_initial_cands(int nc, dm_seq_t *x, dm_seq_t *y);
  /* Generates a list of {nc} candidates for the two {dm_seq_t}s.
    At east one candidate is near the diagonal. */
  
int main(int argc, char**argv)
  { 
    dm_options_t *o = dm_get_options(argc, argv);
    int nc = o->nCands;
    
    fprintf(stderr, "generating the ancestral sequence ...\n");
    char *borg = dm_nucleic_string_throw(o->nBases, TRUE, TRUE);
    
    fprintf(stderr, "writing out the ancestral sequence ...\n");
    dm_nucleic_string_write_named(borg, "Ancestral", o->outName, "-org");
    
    fprintf(stderr, "generating and writing the mutated copies ...\n");
    dm_seq_t x = dm_test_tools_make_seq(0, "X", borg, o->mutProb, o->delProb, o->circX, o->outName);
    dm_seq_t y = dm_test_tools_make_seq(1, "Y", borg, o->mutProb, o->delProb, o->circY, o->outName);
    
    /* Sequence lengths and circularities: */
    int nx = dm_seq_length(&x); bool_t circx = dm_seq_is_circular(&x);
    int ny = dm_seq_length(&y); bool_t circy = dm_seq_is_circular(&y);

    auto double step_score(dm_rung_t g, dm_rung_t h, dm_seq_t *xp, dm_seq_t *yp);

    double step_score(dm_rung_t g, dm_rung_t h, dm_seq_t *xp, dm_seq_t *yp)
      { return dm_score_step(&(o->sc), g, h, xp, yp); }

    fprintf(stderr, "plotting distance matrix for test seq pair ...\n");
    dm_pairing_seq_seq_image_write_named(&x, &y, step_score, o->outName, "-eq");
    
    fprintf(stderr, "creating initial candidates ...\n");
    dm_cand_vec_t cdvraw = dm_fake_initial_cands(nc, &x, &y);

    fprintf(stderr, "writing initial candidates ...\n");
    dm_cand_vec_write_named(&cdvraw, o->outName, "-ini-cd");
    dm_cand_vec_image_write_named(&cdvraw, nx, circx, ny, circy, o->outName, "-ini-cd");
      
    fprintf(stderr, "refining candidates ...\n");
    dm_dyn_tableau_t tb = dm_dyn_tableau_new(); /* Dynamic programming tableau. */
    dm_cand_vec_t cdvref = dm_cand_vec_new(nc); /* Refined cands will be {cdvref[0..nc-1]}. */
    int ic;
    for (ic = 0; ic < nc; ic++)
      { dm_cand_t cd = cdvraw.el[ic];
        /* Refine candidate {o->repeat} times: */
        int t;
        for (t = 0; t < o->repeat; t++)
          { dm_cand_t cdref = dm_refine_cand
              ( &cd, &x, &y,
                o->delta, o->kappa,
                /*shrink*/ TRUE,
                &step_score,
                &tb
              );
            /* Display the matrix: */
            char *tag = NULL;
            asprintf(&tag, "-%05d-%d", ic, t);
            int ngref = dm_pairing_num_rungs(cdref.pr);
            dm_rung_t gopt = dm_pairing_get_rung(cdref.pr, ngref-1);
            dm_dyn_tableau_image_write_named(nx, ny, &tb, gopt, o->outName, tag);
            free(tag);
            /* Prepare for next iteration: */
            cd = cdref;
          }
        cdvref.el[ic] = cd;
      }
      
    fprintf(stderr, "writing final candidates ...\n");
    dm_cand_vec_write_named(&cdvref, o->outName, "-fin-cd");
    dm_cand_vec_image_write_named(&cdvref, nx, circx, ny, circy, o->outName, "-fin-cd");

    return 0;
  }

dm_cand_vec_t dm_fake_initial_cands(int nc, dm_seq_t *x, dm_seq_t *y)
  { /* Sequence lengths and circularities: */
    int nx = dm_seq_length(x); bool_t circx = dm_seq_is_circular(x);
    int ny = dm_seq_length(y); bool_t circy = dm_seq_is_circular(y);
    /* Define the max candidate length {maxlen}: */
    int maxlen = (nx > ny ? nx : ny);
    int minlen = (maxlen + 9)/10;
    /* Generate the candidates: */
    dm_cand_vec_t cdv = dm_cand_vec_throw
      ( nc,
        x->id, x->name, nx, circx, 
        y->id, y->name, ny, circy,
        minlen, maxlen,
        /*circProb*/ 0.1, 
        /*diagProb*/ 0.2,
        /*skipProb*/ 0.1
      );
    return cdv;
  }

dm_options_t *dm_get_options(int argc, char**argv)
  { 
    dm_options_t *o = (dm_options_t *)notnull(malloc(sizeof(dm_options_t)), "no mem");
    
    argparser_t *pp = argparser_new(stderr, argc, argv);
    argparser_set_help(pp, PROG_HELP);
    argparser_set_info(pp, PROG_INFO);
    argparser_process_help_info_options(pp);
    
    argparser_get_keyword(pp, "-nBases");
    o->nBases = argparser_get_next_int(pp, 1, INT_MAX);
    
    if (argparser_keyword_present(pp, "-circular"))
      { o->circX = argparser_get_next_bool(pp);
        o->circY = argparser_get_next_bool(pp);
      }
    else
      { o->circX = o->circY = FALSE; }
    
    argparser_get_keyword(pp, "-mutProb");
    o->mutProb = argparser_get_next_double(pp, 0.0, 1.0);
    
    argparser_get_keyword(pp, "-delProb");
    o->delProb = argparser_get_next_double(pp, 0.0, 1.0 - o->mutProb);
    
    argparser_get_keyword(pp, "-scores");
    o->sc = dm_score_args_parse(pp);
    
    if (argparser_keyword_present(pp, "-delta"))
      { o->delta = argparser_get_next_int(pp, 0, INT_MAX); }
    else
      { o->delta = 3; }
    
    if (argparser_keyword_present(pp, "-kappa"))
      { o->kappa = argparser_get_next_int(pp, 0, INT_MAX); }
    else
      { o->kappa = 6; }
    
    argparser_get_keyword(pp, "-nCands");
    o->nCands = argparser_get_next_int(pp, 0, INT_MAX);
    
    if (argparser_keyword_present(pp, "-repeat"))
      { o->repeat = argparser_get_next_int(pp, 0, INT_MAX); }
    else
      { o->repeat = 1; }
    
    argparser_skip_parsed(pp);
    
    o->outName = argparser_get_next(pp);

    argparser_finish(pp);
    
    return o;
  }
