/* See {dm_test_tools.h} */
/* Last edited on 2007-01-04 02:40:09 by stolfi */

#define dm_psplot_C_COPYRIGHT \
  "Copyright  2006  by the State University of Campinas (UNICAMP)"

#include <dm_test_tools.h>

#include <dm_basic.h>
#include <dm_seq.h>
#include <dm_nucleic.h>
#include <dm_refine.h>

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

void dm_test_tools_write_generated_sequence(char *b, dm_seq_t *s, char *outName);
  /* Writes the nucleotide string {b} and its numerically encoded 
    version {s} to disk files named "{outName}-{s.name}.bas" 
    and "{outName}-{s.name}.egs", respectively. */

dm_seq_t dm_test_tools_make_seq
  ( dm_seq_id_t id, 
    char *tag, 
    char *borg,
    double mutProb, 
    double delProb, 
    bool_t circ,
    char *outName
  )
  { char *bmut;
    dm_rung_vec_t gv;
    dm_nucleic_string_mutate(borg, TRUE, TRUE, mutProb, delProb, &bmut, &gv);
    /* Generate comment: */
    char *bcmt = NULL;
    asprintf(&bcmt, "Mutated mp = %8.6f dp = %8.6f\n", mutProb, delProb);
    dm_seq_t s = dm_seq_from_nucleic_string(id, tag, bcmt, bmut, circ);
    /* Write sequence to disk: */
    dm_test_tools_write_generated_sequence(bmut, &s, outName);
    free(bmut); free(gv.el);
    return s;
  }
  
#define dm_rung_huge (dm_rung_t){{ INT_MAX, INT_MAX }}

void dm_test_tools_make_seq_pair
  ( char *borg, 
    double mutProb, 
    double delProb, 
    bool_t circ,
    dm_seq_id_t xid, 
    char *xtag,
    dm_seq_t *xP,
    dm_seq_id_t yid, 
    char *ytag,
    dm_seq_t *yP,
    dm_rung_vec_t *gv,
    char *outName
  )
  { /* Generate two mutated copies of {b}, save the indices of copied chars: */
    char *xdna, *ydna; /* The mutated copies. */
    dm_rung_vec_t xgv, ygv;  /* Index pairs for letters that were copied. */
    dm_nucleic_string_mutate(borg, TRUE, TRUE, mutProb, delProb, &xdna, &xgv);
    dm_nucleic_string_mutate(borg, TRUE, TRUE, mutProb, delProb, &ydna, &ygv);
    /* Merge the rung vectors {xgv,ygv} by component 0 to create the rung vector {*gv}: */
    int ngmax = (xgv.nel < ygv.nel ? xgv.nel : ygv.nel);
    (*gv) = dm_rung_vec_new(ngmax);
    int ng = 0; /* Output rungs will be {gv[0..ng-1]}. */
    int kx = 0, ky = 0;
    while ((kx < xgv.nel) && (ky < ygv.nel))
      { dm_rung_t xg = (kx >= xgv.nel ? dm_rung_huge : xgv.el[kx]);
        dm_rung_t yg = (ky >= ygv.nel ? dm_rung_huge : ygv.el[ky]);
        if (xg.c[0] < yg.c[0]) 
          { kx++; }
        else if (xg.c[0] > yg.c[0])
          { ky++; }
        else
          { dm_rung_t g = (dm_rung_t){{ xg.c[1], yg.c[1] }};
            if (ng > 0) { (void)dm_rung_step_is_monotonic(&(gv->el[ng-1]), &g, /*die*/ TRUE); }
            dm_rung_vec_expand(gv, ng);
            gv->el[ng] = g; ng++; 
            kx++; ky++; 
          }
      }
    dm_rung_vec_trim(gv, ng);
    /* Generate comment: */
    char *cmt = NULL;
    asprintf(&cmt, "Mutated mp = %8.6f dp = %8.6f\n", mutProb, delProb);
    (*xP) = dm_seq_from_nucleic_string(xid, xtag, cmt, xdna, circ);
    (*yP) = dm_seq_from_nucleic_string(yid, ytag, cmt, ydna, circ);
    /* Write sequences to disk: */
    dm_test_tools_write_generated_sequence(xdna, xP, outName);
    dm_test_tools_write_generated_sequence(ydna, yP, outName);
    free(xdna); free(xgv.el);
    free(ydna); free(ygv.el);
  }

void dm_test_tools_write_generated_sequence(char *b, dm_seq_t *s, char *outName)
  { /* Write sequence to disk: */
    char *fileTag = NULL;
    asprintf(&fileTag, "-%s", s->name);
    dm_nucleic_string_write_named(b, s->cmt, outName, fileTag);
    dm_seq_write_named(s, outName, fileTag);
    free(fileTag);
  }
  
void dm_test_tools_write_and_plot_multi_seq(dm_seq_t r[], int maxLevel, char *outName)
  {
    int level;
    for (level = 0; level <= maxLevel; level++)
      { int nd = dm_seq_length(&(r[level]));
        fprintf(stderr, "level %d sequence has %d samples\n", level, nd);
        char *tag = NULL;
        asprintf(&tag, "-%02d", level);
        dm_seq_write_named(&(r[level]), outName, tag);
        dm_seq_postscript_plot_named(&(r[level]), 190.0, 30.0, 1.0, outName, tag);
        free(tag);
      }
  }

dm_cand_t dm_test_tools_get_optimum_pairing
  ( dm_seq_t *xp, 
    dm_seq_t *yp, 
    int delta, 
    dm_step_score_proc_t *step_score, 
    dm_dyn_tableau_t *tb
  ) 
  { /* Get some parameters: */
    int nx = dm_seq_length(xp);
    int ny = dm_seq_length(yp);
    int nmax = (nx > ny ? nx : ny);
    /* Build a Bresenham pairing {gv[0..ng-1]} betwen {xp} and {yp}: */
    int ng = 0; 
    dm_rung_vec_t gv = dm_rung_vec_new(nmax);
    /* Store the initial rung {gini}: */
    dm_rung_t gini = (dm_rung_t){{ 0, 0 }};
    dm_rung_vec_expand(&gv,ng); gv.el[ng] = gini; ng++;
    /* Apped a Bresenham path from {gini} to the final rung {gfin}: */
    dm_rung_t gfin = (dm_rung_t){{ nx-1, ny-1 }};
    dm_rung_bresenham(gini, gfin, &ng, &gv);
    dm_rung_vec_trim(&gv, ng);
    /* Make it into a candidate: */
    dm_pairing_t *pr = dm_pairing_from_rung_vec(&gv, FALSE);
    dm_cand_t cdraw = dm_cand_from_pairing(xp, yp, pr, /*score*/ 0.0);
    /* Refine it: */
    dm_cand_t cdopt = dm_refine_cand(&cdraw, xp, yp, delta, 0, FALSE, step_score, tb);
    return cdopt; 
  }
